# -*- coding: utf-8 -*-

"""Package where the BaseHyperParam class is defined."""

from abc import ABCMeta, abstractmethod

import numpy as np


class BaseHyperParam(metaclass=ABCMeta):
    """BaseHyperParam class.

    A BaseHyperParam is an abstract representation of a single parameter that can be tuned.

    Attributes:
        dimensions (int):
            Number of dimensions that the hyperparameter uses to be represented in the search
            space.
        cardinality (int or np.inf):
            Number of possible values for this hyperparameter.
    """

    dimensions = 0
    cardinality = 0
    default = None

    def _to_array(self, values):
        """Validate and convert ``values`` to a ``numpy.ndarray`` with the expected shape.

        If ``self.dimensions`` is 1, only scalars, or lists of scalars or 1d arrays or
        2d arrays with one column are accepted and are transformed to:

            - scalar: 2d array with a single row and column.

            - list of scalars: 2d array with the list of scalars as its single column.

            - 1d array: 2d array with the given 1d array as its only column.

            - 2d array: the same array is returned unmodified.

        If ``self.dimensions`` is greater than one, only lists of scalars, or lists of lists
        or 1d arrays or 2d arrays are accepted and are transformed to:

            - list of scalars: 2d array with the list of scalars as its single row.

            - 1d array: 2d array with the given 1d array as its only row.

            - 2d array: the same array is returned unmodified.

        Args:
            values (scalar or array-like):
                A single scalar value or array-like of values to be converted to ``numpy.array``.

        Returns:
            numpy.ndarray:
                Values converted into ``numpy.ndarray`` with shape ``(, self.dimensions)``.

        Raises:
            ValueError:
                If the given values cannot fit into the expected output shape.
        """

        if np.isscalar(values):
            if self.dimensions == 1:
                return np.array([[values]])
            else:
                raise ValueError(
                    "Only lists or numpy.ndarrays are supported for dimensions > 1"
                )

        if isinstance(values, list):
            if not all(np.isscalar(value) for value in values):
                if not all(isinstance(value, list) for value in values):
                    raise ValueError("Only list of lists are supported")
                elif not all(len(value) == self.dimensions for value in values):
                    raise ValueError("All sublists must have len == dimensions")

            values = np.array(values)

        if len(values.shape) > 2:
            raise ValueError("Invalid shape: Too many dimensions")

        if self.dimensions == 1:
            if len(values.shape) == 1:
                values = values.reshape(-1, 1)
            elif values.shape[1] != 1:
                raise ValueError(
                    "Invalid shape: Only 1 column is supported if dimensions == 1"
                )

        elif len(values.shape) == 1:
            if len(values) != self.dimensions:
                raise ValueError("Number of elements != number of dimensions")
            elif not all(np.isscalar(value) for value in values):
                raise ValueError("Numpy arrays must only contain scalars")

            values = values.reshape(1, -1)

        return values

    def _within_range(self, values, min=0, max=1):
        """Ensure that the values are within a certain range.

        Args:
            values (numpy.ndarray):
                2D array of values that will be validated.

        Raises:
            ValueError:
                A ``ValueError`` is raised if any value from ``values`` is not within the range.
        """
        inside_mask = np.ma.masked_inside(values, min, max)
        if inside_mask.any():
            outside = inside_mask[~inside_mask.mask].data.reshape(-1).tolist()
            raise ValueError(
                "Values found outside of the valid range [{}, {}]: {}".format(
                    min, max, outside
                )
            )

    def _within_hyperparam_space(self, values):
        """Ensure that the values are within the range of the hyperparameter space.

        Args:
            values (numpy.ndarray):
                2D array of values that will be validated.
        """
        self._within_range(values, min=self.min, max=self.max)

    def _within_search_space(self, values):
        """Ensure that the values are within the range of the search space.

        Args:
            values (numpy.ndarray):
                2D array of values that will be validated.
        """
        self._within_range(values.astype(np.float64), min=0, max=1)

    @abstractmethod
    def _inverse_transform(self, values):
        """Method to be implemented by child classes."""
        pass

    @abstractmethod
    def _transform(self, values):
        """Method to be implemented by child classes."""
        pass

    def inverse_transform(self, values):
        """Invert one or more search space values.

        Validates that the input values are within the search space and then transform them into
        hyperparameter values.

        Args:
            values (scalar or array-like):
                Scalar or array-like of values to be converted into the hyperparameter space.

        Returns:
            numpy.ndarray:
                2D ``numpy.ndarray`` containing values from the original hyperparameter space.

        Example:
            The example below shows simple usage case where an ``IntHyperParam`` is being imported,
            instantiated with a range from 1 to 4, and its method ``inverse_transform`` is being
            called two times with a scalar from the search space and an array of two valid values
            from the search space.

            >>> from baytune.tuning.hyperparams.numerical import IntHyperParam
            >>> ihp = IntHyperParam(min=1, max=4)
            >>> ihp.inverse_transform(0.125)
            array([[1]])
            >>> ihp.inverse_transform([0.125, 0.375])
            array([[1],
                   [2]])
        """
        values = self._to_array(values)
        self._within_search_space(values)

        return self._inverse_transform(values)

    @abstractmethod
    def sample(self, n_samples):
        """Generate an array of ``n_samples`` random samples in the search space.

        Args:
            n_samples (int):
                Number of values to sample.

        Returns:
            numpy.ndarray:
                2D ``numpy.ndarray`` with a shape `(n_samples, self.dimensions)`.

        Example:
            The example below shows simple usage case where an ``IntHyperParam`` is being imported,
            instantiated with a range from 1 to 4, and its method ``sample`` is being called
            with a number of samples to be obtained. A ``numpy.ndarray`` with values from the
            search space is being returned.

            >>> from baytune.tuning.hyperparams.numerical import IntHyperParam
            >>> instance = IntHyperParam(min=1, max=4)
            >>> instance.sample(2)
            array([[0.625],
                   [0.375]])
        """

    def transform(self, values):
        """Transform one or more hyperparameter values.

        Validates that the input values are within the accepted dimensions and that they are within
        the hyperparameter space. Then transform one or more hyperparameter values from the
        original hyperparameter space into the normalized search space :math:`[0, 1]^K`.
        The accepted value formats are:

            - Scalar:
                A single scalar value from the original hyperparameter space.
            - List:
                A list composed by values from the original hyperparameter space.
            - 2D array-like:
                Two dimensions array-like object that contains values from the original
                hyperparameter space.

        Args:
            values (scalar, list or array-like):
                Single scalar value, list of values or array-like of values from the
                hyperparameter space to be converted into the search space.

        Returns:
            numpy.ndarray:
                2D ``numpy.ndarray`` of shape `(len(values), self.dimensions)` containing the
                search space values.

        Example:
            The example below shows simple usage case where an ``IntHyperParam`` is being imported,
            instantiated with a range from 1 to 4, and its method ``transform`` is being called
            three times with a single scalar value, an array of two valid values and a 2D array
            with 1 dimension.

            >>> from baytune.tuning.hyperparams.numerical import IntHyperParam
            >>> ihp = IntHyperParam(min=1, max=4)
            >>> ihp.transform(1)
            array([[0.125]])
            >>> ihp.transform([1, 2])
            array([[0.125],
                   [0.375]])
            >>> ihp.transform([[1], [2]])
            array([[0.125],
                   [0.375]])
        """

        if not isinstance(values, np.ndarray):
            values = np.asarray(values)

        dimensions = len(values.shape)
        if dimensions > 2:
            raise ValueError("Too many dimensions.")

        elif dimensions < 2:
            values = values.reshape(-1, 1)

        if values.shape[1] > 1:
            raise ValueError("Only one column is supported.")

        self._within_hyperparam_space(values)

        return self._transform(values)



# -*- coding: utf-8 -*-

import numpy as np

from btb_benchmark.challenges.challenge import Challenge

B = 5.1 / (4 * pow(np.pi, 2))
C = 5 / np.pi
T = 1 / (8 * np.pi)


class Branin(Challenge):
    r"""Branin challenge.

    The Branin, or Branin-Hoo, function is commonly used as a test function for metamodeling
    in computer experiments, especially in the context of optimization. This function takes as
    input ``x`` and ``y`` and has six constants that are named as ``a, b, c, r, s, t``.

    Reference:
        https://uqworld.org/t/branin-function/53https://uqworld.org/t/branin-function/53

    The function is defined by:
       :math:`f(x, y)=\left(y-\frac{5.1 x^2_1}{4 \pi^2}+\frac{5 x}{\pi}-6\right)^2+10\left(1-
       \frac{1}{8 \pi}\right) \cos(x)+10`

    It has a global minimum, with the default `a, b, c, r, s, t` at the following three points:
        :math:`f(x, y) = 0.397887` at :math:`x, y = (-\pi, 12.275), (\pi, 2.275)` and
        :math:`(9.42478, 2.475)`

    Args:
        a (int):
            Constant value for ``a``. Defaults to 1.
        b (float or int):
            Constant value for ``b``. Defaults to :math:`5.1 / (4*\pi^2)`.
        c (float or int):
            Constant value for ``c``. Defaults to :math:`5 / \pi`.
        r (int):
            Constant value for ``r``. Defaults to 6
        s (int):
            Constant value for ``s``. Defaults to 10
        t (float or int):
            Constant value for ``t``. Defaults to :math:`1 / 8*\pi`.
    """
    def __init__(self, a=1, b=B, c=C, r=6, s=10, t=T,
                 min_x=-5.0, max_x=10.0, min_y=0.0, max_y=15.0):
        self.a = a
        self.b = b
        self.c = c
        self.r = r
        self.s = s
        self.t = t
        self.min_x = min_x
        self.max_x = max_x
        self.min_y = min_y
        self.max_y = max_y

    def get_tunable_hyperparameters(self):
        return {
            'x': {
                'type': 'float',
                'range': [self.min_x, self.max_x],
                'default': None
            },
            'y': {
                'type': 'float',
                'range': [self.min_y, self.max_y],
                'default': None
            }
        }

    def evaluate(self, x, y):
        z = (y - self.b * x**2 + self.c * x - self.r)**2
        return -1 * (self.a * z + self.s * (1 - self.t) * np.cos(x) + self.s)



import random

from baytune.selection.selector import Selector


class CustomSelector(Selector):
    """Custom selector"""

    def select(self, choice_scores):
        """Select a choice uniformly at random."""
        return self.choices[random.randint(0, len(self.choices) - 1)]



# -*- coding: utf-8 -*-

from ax import optimize


def convert_hyperparameters(dict_hyperparams):
    """Convert the hyperparameters to the format accepted by the library."""
    hyperparams = []

    if not isinstance(dict_hyperparams, dict):
        raise TypeError('Hyperparams must be a dictionary.')

    for name, hyperparam in dict_hyperparams.items():
        hp_type = hyperparam['type']

        if hp_type == 'int':
            hp_range = hyperparam.get('range') or hyperparam.get('values')
            hp_min = int(min(hp_range))
            hp_max = int(max(hp_range))
            hyperparams.append({
                'name': name,
                'type': 'range',
                'bounds': [hp_min, hp_max]
            })

        elif hp_type == 'float':
            hp_range = hyperparam.get('range') or hyperparam.get('values')
            hp_min = float(min(hp_range))
            hp_max = float(max(hp_range))
            hyperparams.append({
                'name': name,
                'type': 'range',
                'bounds': [hp_min, hp_max]
            })

        elif hp_type == 'bool':
            hyperparams.append({
                'name': name,
                'type': 'choice',
                'bounds': [True, False]
            })

        elif hp_type == 'str':
            hp_range = hyperparam.get('range') or hyperparam.get('values')
            hyperparams.append({
                'name': name,
                'type': 'choice',
                'bounds': hp_range,
            })

    return hyperparams


def adapt_scoring_function(scoring_function):
    """Adapt the scoring function.

    Ax's optimize function calls the scoring function with a dict object, however our challenges
    recieve them as kwargs.
    """
    def adapted_function(params):
        return scoring_function(**params)

    return adapted_function


def ax_optimize(scoring_function, tunable_hyperparameters, iterations):
    """Construct and run a full optimization loop.

    Convert the hyperparameters in to the accepted format by ``Ax`` and adapt the scoring
    function to pass the configuration as kwargs.

    Args:
        scoring_function (function):
            A function that performs scoring over params.
        tunable_hyperparameters (dict):
            A python dict with hyperparameters.
        iterations (int):
            Number of tuning iterations to perform.
    """
    parameters = convert_hyperparameters(tunable_hyperparameters)
    evaluation_function = adapt_scoring_function(scoring_function)

    best_params = optimize(
        parameters=parameters,
        evaluation_function=evaluation_function,
        total_trials=iterations,
        minimize=False
    )[0]

    return evaluation_function(best_params)



# -*- coding: utf-8 -*-

from btb_benchmark.main import run_benchmark

__author__ = """MIT Data To AI Lab"""
__email__ = 'dailabmit@gmail.com'
__version__ = '0.4.1.dev0'
__all__ = ('run_benchmark', )



# -*- coding: utf-8 -*-

from btb_benchmark.tuning_functions.ax import ax_optimize
from btb_benchmark.tuning_functions.btb import (
    gcpeituner, gcptuner, gpeituner, gptuner, uniformtuner)
from btb_benchmark.tuning_functions.hyperopt import hyperopt_tpe
from btb_benchmark.tuning_functions.skopt import skopt_EI, skopt_gp_hedge, skopt_LCB, skopt_PI
from btb_benchmark.tuning_functions.smac import (
    smac_hb4ac, smac_smac4hpo_ei, smac_smac4hpo_lcb, smac_smac4hpo_pi)


def get_all_tuning_functions():
    """Return all the tuning functions ready to use with benchmark."""
    return {
        "Ax.optimize": ax_optimize,
        "baytune.GPTuner": gptuner,
        "baytune.GCPTuner": gcptuner,
        "baytune.GPEiTuner": gpeituner,
        "baytune.GCPEiTuner": gcpeituner,
        "baytune.UniformTuner": uniformtuner,
        "HyperOpt.tpe": hyperopt_tpe,
        "skopt.EI": skopt_EI,
        "skopt.gp_hedge": skopt_gp_hedge,
        "skopt.LCB": skopt_LCB,
        "skopt.PI": skopt_PI,
        "SMAC.HB4AC": smac_hb4ac,
        "SMAC.SMAC4HPO_EI": smac_smac4hpo_ei,
        "SMAC.SMAC4HPO_LCB": smac_smac4hpo_lcb,
        "SMAC.SMAC4HPO_PI": smac_smac4hpo_pi,
    }



# -*- coding: utf-8 -*-

"""Package where the BooleanHyperParam class is defined."""

import numpy as np

from baytune.tuning.hyperparams.base import BaseHyperParam


class BooleanHyperParam(BaseHyperParam):
    """BooleanHyperParam class.

    The BooleanHyperParam class is responsible for the transformation of boolean values in
    to normalized search space of :math:`[0, 1]`, providing the ability to sample values of those
    and to inverse transform from the search space into the hyperparameter space.

    Hyperparameter space:
        ``{True, False}``

    Args:
        default (bool):
            Default boolean value for the hyperparameter. Defaults to ``False``.
    """

    dimensions = 1
    cardinality = 2

    def __init__(self, default=False):
        self.default = default

    def _within_hyperparam_space(self, values):
        if values.dtype is not np.dtype("bool"):
            # values is expected to be np.ndarray(n, 1) [[False], [True]]
            if not all(isinstance(value[0], bool) for value in values):
                raise ValueError(
                    "Values: {} not within hyperparameter space.".format(values)
                )

    def _inverse_transform(self, values):
        """Invert one or more search space values.

        Converts a ``numpy.ndarray`` with normalized values from the search space :math:`{0, 1}`
        to the original space of ``True`` and ``False`` by casting those values to ``boolean``.

        Args:
            values (numpy.ndarray):
                2D array with values from the search space.

        Returns:
            numpy.ndarray:
                2D ``numpy.ndarray`` containing values from the original hyperparameter space.

        Example:
            The example below shows simple usage case where a BooleanHyperParam is being created
            and it's ``_inverse_transform`` method is being called with a valid ``numpy.ndarray``
            containing a valid shape of data.

            >>> instance = BooleanHyperParam()
            >>> instance._inverse_transform(np.array([[0], [1]]))
            array([[ True],
                   [False]])
            >>> instance._inverse_transform(np.array([[0]]))
            array([[False]])
        """
        return values.astype(bool)

    def _transform(self, values):
        r"""Transform one or more hyperparameter values.

        Converts a ``numpy.array`` with values from the original hyperparameter space into
        normalized values in the search space of :math:`\{0, 1\}` by casting those values to
        ``int``.

        Args:
            values (numpy.ndarray):
                2D array with values from the hyperparameter space to be converted into the
                search space.

        Returns:
            numpy.ndarray:
                2D ``numpy.ndarray`` of shape ``(len(values), 1)`` containing the search space
                values.

        Example:
            The example below shows simple usage case where a BooleanHyperParam is being created
            and it's ``_transform`` method is being called with a valid ``numpy.ndarray``.

            >>> instance = BooleanHyperParam()
            >>> instance._transform(np.array([[True], [False]]))
            array([[1],
                   [0]])
            >>> instance._transform(np.array([[True]]))
            array([[1]])
        """
        return values.astype(int)

    def sample(self, n_samples):
        """Generate sample values in the hyperparameter search space :math:`{0, 1}`.

        Args:
            n_samples (int):
                Number of values to sample.

        Returns:
            numpy.ndarray:
                2D array with shape of ``(n_samples, 1)`` with normalized values inside the
                search space :math:`{0, 1}`.

        Example:
            The example below shows simple usage case where a BooleanHyperParam is being created
            and it's ``sample`` method is being called with a number of samples to be obtained.

            >>> instance = BooleanHyperParam()
            >>> instance.sample(2)
            array([[1],
                   [1]])
        """
        sampled = np.random.random((n_samples, self.dimensions))

        return np.round(sampled).astype(int)

    def __repr__(self):
        return "BooleanHyperParam(default={})".format(self.default)



# -*- coding: utf-8 -*-

"""Package where the ExpectedImprovementAcquisition class is defined."""

import numpy as np
from scipy.stats import norm

from baytune.tuning.acquisition.base import BaseAcquisition


class ExpectedImprovementAcquisition(BaseAcquisition):
    def _acquire(self, candidates, num_candidates=1):
        Phi = norm.cdf
        N = norm.pdf

        mu, sigma = candidates.T
        y_best = np.max(self.scores)

        z = (mu - y_best) / sigma

        ei = sigma * (z * Phi(z) + N(z))

        return self._get_max_candidates(ei, num_candidates)



# -*- coding: utf-8 -*-

"""Package where the Challenge class is defined."""

import inspect
from abc import ABCMeta, abstractmethod


class Challenge(metaclass=ABCMeta):
    """Challenge class.

    The Challenge class represents a single ``challenge`` that can be used for benchmark.
    """

    @abstractmethod
    def get_tunable_hyperparameters(self):
        """Return a dictionary with hyperparameters to be tuned."""
        pass

    @abstractmethod
    def evaluate(self, *args, **kwargs):
        """Perform evaluation for the given ``arguments``.

        This method will score a result with a given configuration, then return the score obtained
        for those ``arguments``.
        """
        pass

    def __repr__(self):
        args = inspect.getargspec(self.__init__)
        keys = args.args[1:]
        defaults = dict(zip(keys, args.defaults))
        instanced = {key: getattr(self, key) for key in keys}

        if defaults == instanced:
            return '{}()'.format(self.__class__.__name__)

        else:
            args = ', '.join(
                '{}={}'.format(key, value)
                for key, value in instanced.items()
            )

            return '{}({})'.format(self.__class__.__name__, args)



# -*- coding: utf-8 -*-

"""Package where all the available tuners are imported."""

from baytune.tuning.tuners.gaussian_process import (
    GCPEiTuner,
    GCPTuner,
    GPEiTuner,
    GPTuner,
)
from baytune.tuning.tuners.uniform import UniformTuner

__all__ = (
    "GCPEiTuner",
    "GCPTuner",
    "GPEiTuner",
    "GPTuner",
    "UniformTuner",
)



import os

import pandas as pd


def load_results(files):
    version_results = dict()
    for filename in files:
        version = os.path.basename(filename).replace('.csv', '')
        version_results[version] = pd.read_csv(filename, index_col=0).round(6)

    return version_results


def get_wins(scores):
    is_winner = scores.T.rank(method='min', ascending=False) == 1
    return is_winner.sum(axis=1)


def get_exclusive_wins(scores):
    is_winner = scores.T.rank(method='min', ascending=False) == 1
    num_winners = is_winner.sum()
    is_exclusive = num_winners == 1
    is_exclusive_winner = is_winner & is_exclusive
    num_exclusive_wins = is_exclusive_winner.sum(axis=1)

    return num_exclusive_wins


def get_z_scores(scores):
    mean = scores.mean(axis=1)
    std = scores.std(axis=1)

    return ((scores.T - mean) / std).fillna(0).T.mean()


def get_summary(results, summary_function):
    summary = {}
    for version, scores in results.items():
        summary[version] = summary_function(scores)

    summary_df = pd.DataFrame(summary)
    summary_df.index.name = 'tuner'
    columns = summary_df.columns.sort_values(ascending=False)
    return summary_df[columns]


def add_sheet(dfs, name, writer, cell_fmt, index_fmt, header_fmt):
    startrow = 0
    widths = [0]
    if not isinstance(dfs, dict):
        dfs = {None: dfs}

    for df_name, df in dfs.items():
        df = df.reset_index()
        startrow += bool(df_name)
        df.to_excel(writer, sheet_name=name, startrow=startrow + 1, index=False, header=False)

        worksheet = writer.sheets[name]

        if df_name:
            worksheet.write(startrow - 1, 0, df_name, index_fmt)
            widths[0] = max(widths[0], len(df_name))

        for idx, column in enumerate(df.columns):
            worksheet.write(startrow, idx, column, header_fmt)
            width = max(len(column), *df[column].astype(str).str.len()) + 1
            if len(widths) > idx:
                widths[idx] = max(widths[idx], width)
            else:
                widths.append(width)

        startrow += len(df) + 2

    for idx, width in enumerate(widths):
        fmt = cell_fmt if idx else index_fmt
        worksheet.set_column(idx, idx, width + 1, fmt)


def write_results(results, output):
    writer = pd.ExcelWriter(output, engine='xlsxwriter')
    cell_fmt = writer.book.add_format({
        "font_name": "Arial",
        "font_size": "10"
    })
    index_fmt = writer.book.add_format({
        "font_name": "Arial",
        "font_size": "10",
        "bold": True,
    })
    header_fmt = writer.book.add_format({
        "font_name": "Arial",
        "font_size": "10",
        "bold": True,
        "bottom": 1
    })

    wins = get_summary(results, get_wins)
    summary = {
        'Number of Wins (with ties)': wins,
        'Number of Wins (exclusive)': get_summary(results, get_exclusive_wins)
    }
    add_sheet(summary, 'Number of Wins', writer, cell_fmt, index_fmt, header_fmt)

    for version in wins.columns:
        add_sheet(results[version], version, writer, cell_fmt, index_fmt, header_fmt)

    writer.save()



from baytune.selection.ucb1 import UCB1


class HierarchicalByAlgorithm(UCB1):
    """Hierarchical selector

    Args:
        by_algorithm (Dict[str, List]): mapping of ML algorithms to frozen set choices
    """

    def __init__(self, choices, by_algorithm):
        super(HierarchicalByAlgorithm, self).__init__(choices)
        self.by_algorithm = by_algorithm

    def select(self, choice_scores):
        """
        Groups the frozen sets by algorithm and first chooses an algorithm based
        on the traditional UCB1 criteria.

        Next, from that algorithm's frozen sets, makes the final set choice.
        """
        # choose algorithm using a bandit
        alg_scores = {}
        for algorithm, choices in self.by_algorithm.items():
            # only make arms for algorithms that have options
            if not set(choices) & set(choice_scores.keys()):
                continue
            # sum up lists to get a list of all the scores from any run of this
            # algorithm
            sublists = [choice_scores.get(c, []) for c in choices]
            alg_scores[algorithm] = sum(sublists, [])

        best_algorithm = self.bandit(alg_scores)

        # now use only the frozen sets from the chosen algorithm
        best_subset = self.by_algorithm[best_algorithm]
        normal_ucb1 = UCB1(choices=best_subset)
        return normal_ucb1.select(choice_scores)



# -*- coding: utf-8 -*-

"""Package where the CategoricalHyperParamClass is defined."""

from copy import deepcopy

import numpy as np
from sklearn.preprocessing import OneHotEncoder

from baytune.tuning.hyperparams.base import BaseHyperParam


class CategoricalHyperParam(BaseHyperParam):
    r"""CategoricalHyperParam Class.

    The CategoricalHyperParam class is responsible for the transform of categorical values
    in to normalized search space and provides the inverse transform from search space to
    hyperparameter space. Also provides a method that generates samples of those.

    Hyperparameter space:
        :math:`h_1, h_2,... h_K` where `K` is the number of categories.

    Search Space:
        :math:`\{ 0, 1 \}^K` where `K` is the number of categories.

    Args:
        choices (list):
            List of values that the hyperparameter can be.

        default (str or None):
            Default value for the hyperparameter to take. Defaults to the first item in ``choices``
    """
    NO_DEFAULT = object()

    def __init__(self, choices, default=NO_DEFAULT):
        """Instantiation of CategoricalHyperParam.

        Creates an instance with a list of ``choices`` and fit an instance of
        ``sklearn.preprocessing.OneHotEncoder`` with those values.
        """
        if default is self.NO_DEFAULT:
            self.default = choices[0]
        elif default not in choices:
            raise ValueError("`default` not within `choices`")
        else:
            self.default = default

        self.choices = deepcopy(choices)
        self.dimensions = len(choices)
        self.cardinality = self.dimensions
        choices = np.array(choices, dtype="object")
        self._encoder = OneHotEncoder(categories=[choices], sparse_output=False)
        self._encoder.fit(choices.reshape(-1, 1))

    def _within_hyperparam_space(self, values):
        mask = np.isin(values, self.choices)

        if not mask.all():
            if not isinstance(values, np.ndarray):
                values = np.asarray(values)

            not_in_space = values[~mask].tolist()
            raise ValueError(
                "Values found outside of the valid space {}: {}".format(
                    self.choices, not_in_space
                )
            )

    def _inverse_transform(self, values):
        """Invert one or more values.

        Converts a ``numpy.ndarray`` with normalized values from the search space to the original
        hyperparameter space of ``self.choices``.

        Args:
            values (numpy.ndarray):
                2D array with values from the search space.

        Returns:
            numpy.ndarray:
                2D ``numpy.ndarray`` containing values from the original hyperparameter space.

        Example:
            The example below shows simple usage case where a CategoricalHyperParam is being
            created with three possible values, (Cat, Dog, Tiger), and it's method
            ``_inverse_transform`` is being called with a valid ``numpy.ndarray`` containing
            values from the search space and values from the hyperparameter space are being
            returned.

            >>> instance = CategoricalHyperParam(choices=['Cat', 'Dog', 'Tiger'])
            >>> instance._inverse_transform(np.array([[1, 0, 0]]))
            array([['Cat']])
            >>> instance._inverse_transform(np.array([[1, 0, 0], [0, 0, 1]]))
            array([['Cat'],
                   ['Tiger']])
        """
        if len(values.shape) == 1:
            values = values.reshape(1, -1)

        return self._encoder.inverse_transform(values.astype("object"))

    def _transform(self, values):
        """Transform one or more categorical values.

        Encodes one or more categorical values in to the normalized search space of
        :math:`[0, 1]^K` by using ``sklearn.preprocessing.OneHotEncoder`` that has been
        fitted during the instantiation.

        Args:
            values (numpy.ndarray):
                2D array with values from the hyperparameter space to be converted into the
                search space.

        Returns:
            numpy.ndarray:
                2D ``numpy.ndarray`` of shape `(len(values), self.dimensions)` containing the
                search space values.

        Example:
            The example below shows simple usage case where a CategoricalHyperParam is being
            created with three possible values, (Cat, Dog, Tiger), and it's method ``_transform``
            is being called with a valid ``numpy.ndarray`` containing values from the
            hyperparameter space and an array with normalized values is being returned.

            >>> instance = CategoricalHyperParam(choices=['Cat', 'Dog', 'Tiger'])
            >>> instance._transform(np.array([['Cat']]))
            array([[1, 0, 0]])
            >>> instance._transform(np.array([['Cat'], ['Tiger']]))
            array([[1, 0, 0],
                   [0, 0, 1]])
        """
        return self._encoder.transform(values.astype("object")).astype(int)

    def sample(self, n_samples):
        """Generate sample values in the hyperparameter search space of ``[0, 1]^K``.

        Args:
            n_samples (int):
                Number of values to sample.

        Returns:
            numpy.ndarray:
                2D array with shape of `(n_samples, self.dimensions)` with normalized values
                inside the search space :math:`[0, 1]^K`.

        Example:
            The example below shows simple usage case where a CategoricalHyperParam is being
            created with three possible values, (Cat, Dog, Tiger), and it's method ``sample``
            is being called with a number of samples to be obtained. A ``numpy.ndarray`` with
            values from the search space is being returned.

            >>> instance = CategoricalHyperParam(choices=['Cat', 'Dog', 'Tiger'])
            >>> instance.sample(2)
            array([[1, 0, 0],
                   [0, 1, 0]])
        """
        randomized_values = np.random.random((n_samples, self.dimensions))
        indexes = np.argmax(randomized_values, axis=1)

        sampled = [self.choices[index] for index in indexes]

        return self.transform(sampled)

    def __repr__(self):
        return "CategoricalHyperParam(choices={}, default={})".format(
            self.choices, self.default
        )



# -*- coding: utf-8 -*-

import numpy as np

from baytune.tuning.tunable import Tunable
from baytune.tuning.tuners import GCPEiTuner, GCPTuner, GPEiTuner, GPTuner, UniformTuner


def _tuning_function(tuner_class, scoring_function, tunable_hyperparameters, iterations):
    tunable = Tunable.from_dict(tunable_hyperparameters)
    tuner = tuner_class(tunable)
    best_score = -np.inf

    for _ in range(iterations):
        proposal = tuner.propose()
        score = scoring_function(**proposal)
        tuner.record(proposal, score)
        best_score = max(score, best_score)

    return best_score


def make_btb_tuning_function(tuner_class):
    """Create a tuning function for a tuner class based on ``BTB``.

    Args:
        tuner_class (baytune.tuning.tuners.base.BaseTuner):
            A tuner class based on the BTB ``BaseTuner``.

    Return:
        callable:
            Return a tuning function using the given tuner class.
    """
    def btb_tuning_function(scoring_function, tunable_hyperparameters, iterations):
        return _tuning_function(tuner_class, scoring_function, tunable_hyperparameters, iterations)

    return btb_tuning_function


def gptuner(scoring_function, tunable_hyperparameters, iterations):
    return _tuning_function(GPTuner, scoring_function, tunable_hyperparameters, iterations)


def gpeituner(scoring_function, tunable_hyperparameters, iterations):
    return _tuning_function(GPEiTuner, scoring_function, tunable_hyperparameters, iterations)


def gcptuner(scoring_function, tunable_hyperparameters, iterations):
    return _tuning_function(GCPTuner, scoring_function, tunable_hyperparameters, iterations)


def gcpeituner(scoring_function, tunable_hyperparameters, iterations):
    return _tuning_function(GCPEiTuner, scoring_function, tunable_hyperparameters, iterations)


def uniformtuner(scoring_function, tunable_hyperparameters, iterations):
    return _tuning_function(UniformTuner, scoring_function, tunable_hyperparameters, iterations)



# -*- coding: utf-8 -*-

import numpy as np
import scipy
from copulas import EPSILON
from copulas.univariate import Univariate
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF

from baytune.tuning.metamodels.base import BaseMetaModel


class GaussianProcessMetaModel(BaseMetaModel):
    """GaussianProcessMetaModel class.

    This class represents a meta-model using an underlying ``GaussianProcessRegressor`` from
    ``sklearn.gaussian_process``.

    Attributes:
        _MODEL_KWARGS (dict):
            Dictionary with the default ``kwargs`` for the ``GaussianProcessRegressor``
            instantiation.
        _MODEL_CLASS (type):
            Class to be instantiated and used for the ``self._model`` instantiation. In
            this case ``sklearn.gaussian_process.GaussainProcessRegressor``
    """

    _MODEL_CLASS = GaussianProcessRegressor

    _MODEL_KWARGS_DEFAULT = {"normalize_y": True}

    def __init_metamodel__(self, length_scale=1):
        if self._model_kwargs is None:
            self._model_kwargs = {}

        self._model_kwargs["kernel"] = RBF(length_scale=length_scale)

    def _predict(self, candidates):
        predictions = self._model_instance.predict(candidates, return_std=True)
        return np.column_stack(predictions)


class GaussianCopulaProcessMetaModel(GaussianProcessMetaModel):
    """GaussianCopulaProcessMetaModel class.

    This class represents a meta-model using an underlying ``GaussianProcessRegressor`` from
    ``sklearn.gaussian_process`` applying ``copulas.univariate.Univariate`` transformations
    to the input data and afterwards reverts it for the predictions.

    During the ``fit`` process, this metamodel trains a univariate copula for each
    hyperparameter to then compute the cumulative distribution of these. Once the cumulative
    distribution has been calculated, we calculate the inverse of the normal cumulative
    distribution using ``scipy.stats.norm`` and use these transformations to train the
    ``GaussianProcessRegressor`` model.

    When predicting the output value, an inverse of the normal cumulative distribution is
    computed to the normal cumulative distribution, using the previously trained univariate
    copula with the input data of the score.

    Attributes:
        _MODEL_KWARGS (dict):
            Dictionary with the default ``kwargs`` for the ``GaussianProcessRegressor``
            instantiation.
        _MODEL_CLASS (type):
            Class to be instantiated and used for the ``self._model`` instantiation. In
            this case ``sklearn.gaussian_process.GaussainProcessRegressor``
    """

    def _transform(self, trials):
        transformed = []
        for column, distribution in zip(trials.T, self._distributions):
            transformed.append(
                scipy.stats.norm.ppf(
                    distribution.cdf(column).clip(0 + EPSILON, 1 - EPSILON)
                )
            )

        return np.column_stack(transformed)

    def _fit(self, trials, scores):
        self._distributions = []
        for column in trials.T:
            distribution = Univariate()
            distribution.fit(column)
            self._distributions.append(distribution)

        distribution = Univariate()
        distribution.fit(scores)
        self._score_distribution = distribution

        trans_trials = self._transform(trials)
        trans_scores = scipy.stats.norm.ppf(
            self._score_distribution.cdf(scores).clip(0 + EPSILON, 1 - EPSILON)
        )

        super()._fit(trans_trials, trans_scores)

    def _predict(self, candidates):
        trans_candidates = self._transform(candidates)
        predicted = super()._predict(trans_candidates)
        cdf = scipy.stats.norm.cdf(predicted)
        if len(cdf.shape) == 2:
            cdf = cdf[:, 0]
        return self._score_distribution.ppf(cdf)



from baytune.selection.best import BestKReward, BestKVelocity
from baytune.selection.hierarchical import HierarchicalByAlgorithm
from baytune.selection.pure import PureBestKVelocity
from baytune.selection.recent import RecentKReward, RecentKVelocity
from baytune.selection.ucb1 import UCB1
from baytune.selection.uniform import Uniform

__all__ = (
    "BestKReward",
    "BestKVelocity",
    "HierarchicalByAlgorithm",
    "PureBestKVelocity",
    "RecentKReward",
    "RecentKVelocity",
    "UCB1",
    "Uniform",
)



# -*- coding: utf-8 -*-

"""Package where the UniformTuner class is defined."""

from baytune.tuning.tuners.base import BaseTuner


class UniformTuner(BaseTuner):
    def _propose(self, num_proposals, allow_duplicates):
        """Generate ``num_proposals`` number of candidates.

        Args:
            num_proposals (int):
                Number of candidates to create.
            allow_duplicates (bool):
                If it's ``False``, the tuner will propose trials that are not recorded. Otherwise
                will generate trials that can be repeated.

        Returns:
            numpy.ndarray:
                It returns ``numpy.ndarray`` with shape
                ``(num_proposals, len(self.tunable.hyperparameters)``.
        """
        return self._sample(num_proposals, allow_duplicates)



# -*- coding: utf-8 -*-
import argparse
import logging
import sys
import warnings

import tabulate

from btb_benchmark.main import run_benchmark, summarize_results


def _run(args):
    # Logger setup
    log_level = (3 - args.verbose) * 10
    fmt = '%(asctime)s - %(process)d - %(levelname)s - %(name)s - %(module)s - %(message)s'
    logging.basicConfig(level=log_level, format=fmt)
    logging.getLogger("botocore").setLevel(logging.ERROR)
    logging.getLogger("hyperopt").setLevel(logging.ERROR)
    logging.getLogger("ax").setLevel(logging.ERROR)
    logging.getLogger("urllib3").setLevel(logging.CRITICAL)

    # run
    results = run_benchmark(
        args.tuners,
        args.challenge_types,
        args.challenges,
        args.sample,
        args.iterations,
        args.max_rows,
        args.output_path,
        args.detailed_output,
    )

    if not args.output_path:
        print(tabulate.tabulate(
            results,
            tablefmt='github',
            headers=results.columns
        ))


def _summary(args):
    summarize_results(args.input, args.output)


def _get_parser():
    parser = argparse.ArgumentParser(description='BTB Benchmark Command Line Interface')
    parser.set_defaults(action=None)
    action = parser.add_subparsers(title='action')
    action.required = True

    # Run action
    run = action.add_parser('run', help='Run the BTB Benchmark')
    run.set_defaults(action=_run)
    run.set_defaults(user=None)

    run.add_argument('-v', '--verbose', action='count', default=0,
                     help='Be verbose. Use -vv for increased verbosity.')
    run.add_argument('-o', '--output-path', type=str, required=False,
                     help='Path to the CSV file where the report will be dumped')
    run.add_argument('-s', '--sample', type=int,
                     help='Run only on a subset of the available datasets of the given size.')
    run.add_argument('-i', '--iterations', type=int, default=100,
                     help='Number of iterations to perform per challenge with each candidate.')
    run.add_argument('-c', '--challenges', nargs='+',
                     help='Challenge/s to be used. Accepts multiple names.')
    run.add_argument('-t', '--tuners', nargs='+',
                     help='Tuner/s to be benchmarked. Accepts multiple names.')
    run.add_argument('-C', '--challenge-types', nargs='+',
                     choices=['math', 'sgd', 'random_forest', 'xgboost'],
                     help='Type of challenge/s to use. Accepts multiple names.')
    run.add_argument('-m', '--max-rows', type=int,
                     help='Max amount of rows to use for each dataset.')
    run.add_argument('-d', '--detailed-output', action='store_true',
                     help='Output a detailed dataset with elapsed times.')

    # Summarize action
    summary = action.add_parser('summary', help='Summarize the BTB Benchmark results')
    summary.set_defaults(action=_summary)
    summary.add_argument('input', nargs='+', help='Input path with results.')
    summary.add_argument('output', help='Output file.')

    return parser


def main():
    warnings.filterwarnings("ignore")

    # Parse args
    parser = _get_parser()
    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(0)

    args = parser.parse_args()

    args.action(args)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Top level where all the hyperparameters are imported."""

from baytune.tuning.hyperparams.boolean import BooleanHyperParam
from baytune.tuning.hyperparams.categorical import CategoricalHyperParam
from baytune.tuning.hyperparams.numerical import FloatHyperParam, IntHyperParam

__all__ = (
    "BooleanHyperParam",
    "CategoricalHyperParam",
    "FloatHyperParam",
    "IntHyperParam",
)



import numpy as np


class Selector(object):
    """Base selector

    Args:
        choices (list): a list of discrete choices from which the selector must choose at every
            call to ``select``.
    """

    def __init__(self, choices):
        self.choices = choices

    def compute_rewards(self, scores):
        """Compute rewards from choice's scores

        Convert a list of scores associated with one choice into a list of rewards. Normally, the
        length of the list will be preserved, even if some of the scores are dropped.
        """
        return list(scores)

    def bandit(self, choice_rewards):
        """Return the choice to take next using multi-armed bandit

        Multi-armed bandit method. Accepts a mapping of choices to rewards which indicate their
        historical performance, and returns the choice that we should make next in order to
        maximize expected reward in the long term.

        The default implementation is to return the arm with the highest average score.

        Args:
            choice_rewards (Dict[object, List[float]]): maps choice IDs to lists of rewards.

        Returns:
            str: the name of the choice to take next.
        """
        return max(choice_rewards, key=lambda a: np.mean(choice_rewards[a]))

    def select(self, choice_scores):
        """Select the next best choice to make

        Args:
            choice_scores (Dict[object, List[float]]): Mapping of choice to list of scores for each
                possible choice. The caller is responsible for making sure each choice that is
                possible at this juncture is represented in the dict, even those with no scores.
                Score lists should be in ascending chronological order, that is, the score from the
                earliest trial should be listed first.

                For example::

                    {
                        1: [0.56, 0.61, 0.33, 0.67],
                        2: [0.25, 0.58],
                        3: [0.60, 0.65, 0.68],
                    }
        """
        choice_rewards = {}
        for choice, scores in choice_scores.items():
            if choice not in self.choices:
                continue

            choice_rewards[choice] = self.compute_rewards(scores)

        return self.bandit(choice_rewards)



# -*- coding: utf-8 -*-

"""Package where the MLChallenge class is defined."""

import logging
import os
from copy import deepcopy
from urllib.parse import urljoin

import pandas as pd
from sklearn.metrics import make_scorer
from sklearn.model_selection import KFold, StratifiedKFold, cross_val_score
from sklearn.preprocessing import OneHotEncoder

from btb_benchmark.challenges.challenge import Challenge
from btb_benchmark.challenges.datasets import get_dataset_names

BASE_DATASET_URL = 'https://atm-data.s3.amazonaws.com/'
BUCKET_NAME = 'atm-data'
LOGGER = logging.getLogger(__name__)


class MLChallenge(Challenge):
    """Machine Learning Challenge class.

    The MLChallenge class rerpresents a single ``machine learning challenge`` that can be used for
    benchmark.

    Args:
        model (class):
            Class of a machine learning estimator.
        dataset (str):
            Name or path to a dataset. If it's a name it will try to read it from
            https://atm-data.s3.amazonaws.com/
        target_column (str):
            Name of the target column in the dataset.
        encode (bool):
            Either or not to encode the dataset using ``sklearn.preprocessing.OneHotEncoder``.
        model_defaults (dict):
            Dictionary with default keyword args for the model instantiation.
        make_binary (bool):
            Either or not to make the target column binary.
        tunable_hyperparameters (dict):
            Dictionary representing the tunable hyperparameters for the challenge.
        metric (callable):
            Metric function. If ``None``, then the estimator's metric function
            will be used in case there is otherwise the default that ``cross_val_score`` function
            offers will be used.
        max_rows (int):
            Maximum number of rows to use from each dataset. If ``None``, or if the
            given number is higher than the number of rows in the dataset, the entire
            dataset is used. Defaults to ``None``.
    """

    _data = None

    @classmethod
    def get_dataset_url(cls, name):
        if not name.endswith('.csv'):
            name = name + '.csv'

        return urljoin(BASE_DATASET_URL, name)

    @classmethod
    def get_all_challenges(cls, challenges='all'):
        """Return a list containing the instance of the datasets available.

        If ``challenges`` is provided, filter by those. Challenges can be a string
        representing the collection desired, all or openml100.

        Args:
            challenges (str or list):
                If string, a collection of datasets will be loaded from the module
                ``btb_benchmark.challenges.datasets``. If it's a list those will be
                the challenges attempted to be loaded.

        Returns:
            loaded_challenges (list):
                A list of ``btb_benchmark.challenges.mlchallenge.MLChallenge`` instances
                with the specified ``challenges``.
        """
        if isinstance(challenges, str):
            datasets = get_dataset_names(challenges)
        else:
            datasets = challenges

        loaded_challenges = []
        for dataset in datasets:
            try:
                loaded_challenges.append(cls(dataset))
                LOGGER.info('Dataset %s loaded', dataset)
            except Exception as ex:
                LOGGER.warn('Dataset: %s could not be loaded. Error: %s', dataset, ex)

        LOGGER.info('%s / %s datasets loaded.', len(loaded_challenges), len(datasets))

        return loaded_challenges

    def load_data(self):
        """Load ``X`` and ``y`` over which to perform fit and evaluate."""
        if os.path.isdir(self.dataset):
            X = pd.read_csv(self.dataset)

        else:
            url = self.get_dataset_url(self.dataset)
            X = pd.read_csv(url)

        if self.max_rows:
            X = X.sample(min(len(X), self.max_rows))

        y = X.pop(self.target_column)

        if self.make_binary:
            y = y.iloc[0] == y

        if self.encode:
            ohe = OneHotEncoder(categories='auto')
            X = ohe.fit_transform(X)

        return X, y

    @property
    def data(self):
        if self._data is None:
            self._data = self.load_data()

        return self._data

    def __init__(self, dataset, model=None, target_column=None, encode=None,
                 tunable_hyperparameters=None, metric=None, model_defaults=None,
                 make_binary=None, stratified=None, cv_splits=5, cv_random_state=42,
                 cv_shuffle=True, metric_args={}, max_rows=None):

        self.model = model or self.MODEL
        self.dataset = dataset or self.DATASET
        self.target_column = target_column or self.TARGET_COLUMN
        self.model_defaults = model_defaults or self.MODEL_DEFAULTS
        self.make_binary = make_binary or self.MAKE_BINARY
        self.tunable_hyperparameters = tunable_hyperparameters or self.TUNABLE_HYPERPARAMETERS
        self.max_rows = max_rows

        if metric:
            self.metric = metric
            self.metric_args = metric_args

        else:
            # Allow to either write a metric method or assign a METRIC function
            self.metric = getattr(self, 'metric', self.__class__.METRIC)
            self.metric_args = getattr(self, 'metric_args', self.__class__.METRIC_ARGS)

        self.stratified = self.STRATIFIED if stratified is None else stratified
        # self.X, self.y = self.load_data()

        self.encode = self.ENCODE if encode is None else encode
        self.scorer = make_scorer(self.metric, **self.metric_args)

        if self.stratified:
            self.cv = StratifiedKFold(
                shuffle=cv_shuffle,
                n_splits=cv_splits,
                random_state=cv_random_state
            )
        else:
            self.cv = KFold(
                shuffle=cv_shuffle,
                n_splits=cv_splits,
                random_state=cv_random_state
            )

    def get_tunable_hyperparameters(self):
        return deepcopy(self.tunable_hyperparameters)

    def evaluate(self, **hyperparams):
        """Apply cross validation to hyperparameter combination.

        Args:
            hyperparams (dict):
                A combination of ``self.tunable_hyperparams``.

        Returns:
            score (float):
                Returns the ``mean`` cross validated score.
        """
        hyperparams.update((self.model_defaults or {}))
        model = self.model(**hyperparams)
        X, y = self.data

        return cross_val_score(model, X, y, cv=self.cv, scoring=self.scorer).mean()

    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, self.dataset)



import logging

from sklearn.metrics import f1_score
from xgboost import XGBClassifier

from btb_benchmark.challenges.mlchallenge import MLChallenge

LOGGER = logging.getLogger(__name__)


class XGBoostChallenge(MLChallenge):

    # TARGET
    TARGET_COLUMN = 'class'
    MAKE_BINARY = False

    # CROSS VALIDATE / SCORER
    METRIC = f1_score
    METRIC_ARGS = {'average': 'macro'}
    ENCODE = True
    STRATIFIED = True

    # MODEL
    MODEL_DEFAULTS = {
        'random_state': 0,
        'n_estimators': 100
    }
    MODEL = XGBClassifier
    TUNABLE_HYPERPARAMETERS = {
        "max_depth": {
            "type": "int",
            "default": 3,
            "range": [
                3,
                10
            ]
        },
        "learning_rate": {
            "type": "float",
            "default": 0.1,
            "range": [
                0,
                1
            ]
        },
        "gamma": {
            "type": "float",
            "default": 0,
            "range": [
                0,
                1
            ]
        },
        "min_child_weight": {
            "type": "int",
            "default": 1,
            "range": [
                1,
                10
            ]
        }
    }



import logging

from baytune.selection.selector import Selector

# the minimum number of scores that each choice must have in order to use best-K
# optimizations. If not all choices meet this threshold, default UCB1 selection
# will be used.
K_MIN = 3

logger = logging.getLogger("btb")


class PureBestKVelocity(Selector):
    """Pure Best K Velocity Selector

    Simply returns the choice with the best best-K velocity.
    """

    def __init__(self, choices, k=K_MIN):
        super(PureBestKVelocity, self).__init__(choices)
        self.k = k

    def compute_rewards(self, scores):
        """
        Compute the "velocity" of (average distance between) the k+1 best
        scores. Return a list with those k velocities padded out with zeros so
        that the count remains the same.
        """
        # get the k + 1 best scores in descending order
        best_scores = sorted(scores, reverse=True)[: self.k + 1]
        velocities = [
            best_scores[i] - best_scores[i + 1] for i in range(len(best_scores) - 1)
        ]

        # pad the list out with zeros to maintain the length of the list
        zeros = (len(scores) - self.k) * [0]
        return velocities + zeros

    def select(self, choice_scores):
        """
        Select the choice with the highest best-K velocity. If any choices
        don't have MIN_K scores yet, return the one with the fewest.
        """
        # if we don't have enough scores to do K-selection, fall back to UCB1
        min_num_scores = min([len(s) for s in choice_scores.values()])
        if min_num_scores >= K_MIN:
            logger.info("PureBestKVelocity: using Pure Best K velocity selection")
            reward_func = self.compute_rewards
        else:
            logger.warning(
                "{klass}: Not enough choices to do K-selection; "
                "returning choice with fewest scores".format(klass=type(self).__name__)
            )
            # reward choices with the fewest scores
            # NOTE: "reward_func = lambda " changed to "def reward_func"
            # as per flake8 suggestions
            # reward_func = lambda s: [1] if len(s) == min_num_scores else [0]

            def reward_func(scores):
                return [1] if len(scores) == min_num_scores else [0]

        choice_rewards = {}
        for choice, scores in choice_scores.items():
            if choice not in self.choices:
                continue
            choice_rewards[choice] = reward_func(scores)

        # the default bandit returns the choice with the highest mean reward
        return self.bandit(choice_rewards)



import logging

from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import f1_score

from btb_benchmark.challenges.mlchallenge import MLChallenge

LOGGER = logging.getLogger(__name__)


class RandomForestChallenge(MLChallenge):

    # TARGET
    TARGET_COLUMN = 'class'
    MAKE_BINARY = False

    # CROSS VALIDATE / SCORER
    METRIC = f1_score
    METRIC_ARGS = {'average': 'macro'}
    ENCODE = True
    STRATIFIED = True

    # MODEL
    MODEL = RandomForestClassifier
    MODEL_DEFAULTS = {
        'random_state': 0,
        'n_estimators': 100,
    }
    TUNABLE_HYPERPARAMETERS = {
        "criterion": {
            "type": "str",
            "default": "gini",
            "values": [
                "entropy",
                "gini"
            ]
        },
        "max_features": {
            "type": "str",
            "default": None,
            "values": [
                None,
                "auto",
                "log2",
                "sqrt"
            ]
        },
        "min_samples_split": {
            "type": "int",
            "default": 2,
            "range": [
                2,
                100
            ]
        },
        "min_samples_leaf": {
            "type": "int",
            "default": 1,
            "range": [
                1,
                100
            ]
        },
        "min_weight_fraction_leaf": {
            "type": "float",
            "default": 0.0,
            "range": [
                0.0,
                0.5
            ]
        },
        "min_impurity_decrease": {
            "type": "float",
            "default": 0.0,
            "range": [
                0.0,
                1000.0
            ]
        },
    }



# -*- coding: utf-8 -*-

"""Top level where all the challenges are imported."""

from btb_benchmark.challenges.bohachevsky import Bohachevsky
from btb_benchmark.challenges.branin import Branin
from btb_benchmark.challenges.randomforest import RandomForestChallenge
from btb_benchmark.challenges.rosenbrock import Rosenbrock
from btb_benchmark.challenges.sgd import SGDChallenge
from btb_benchmark.challenges.xgboost import XGBoostChallenge

__all__ = (
    'Bohachevsky',
    'Branin',
    'RandomForestChallenge',
    'Rosenbrock',
    'SGDChallenge',
    'XGBoostChallenge',
)

MATH_CHALLENGES = {
    'bohachevsky': Bohachevsky,
    'branin': Branin,
    'rosenbrock': Rosenbrock,
}



# -*- coding: utf-8 -*-

import numpy as np

from btb_benchmark.challenges.challenge import Challenge


class Bohachevsky(Challenge):
    r"""Bohachevsky challenge.

    The Bohachevsky functions are bowl shape functions. This function is usually evaluated on the
    input domain :math:`x \epsilon [-100, 100], y \epsilon [-100, 100]`.

    Reference:
        https://www.sfu.ca/~ssurjano/boha.html

    The function is defined by:
        :math:`f(x, y) = x^2 + 2y^2 -0.3cos(3\pi x)-0.4cos(4\pi y)+0.7`

    It has one local minimum at:
        :math:`(x, y) = (0, 0)` where :math:`f(x, y) = 0`.

    Args:
        min_x (int):
            Minimum number that the hyperparameter can propouse for ``x``. Defaults to -100.
        max_x (int):
            Maximum number that the hyperparameter can propouse for ``x``. Defaults to 100.
        min_y (int):
            Minimum number that the hyperparameter can propouse for ``y``. Defaults to -100.
        max_y (int):
            Maximum number that the hyperparameter can propouse for ``y``. Defaults to 100.
    """
    def __init__(self, min_x=-100, max_x=100, min_y=-100, max_y=100):
        self.min_x = min_x
        self.max_x = max_x
        self.min_y = min_y
        self.max_y = max_y

    def get_tunable_hyperparameters(self):
        return {
            'x': {
                'type': 'int',
                'range': [self.min_x, self.max_x],
                'default': None
            },
            'y': {
                'type': 'int',
                'range': [self.min_y, self.max_y],
                'default': None
            }
        }

    def evaluate(self, x, y):
        z = 0.3 * np.cos(3 * np.pi * x)
        return -1 * (x**2 + 2 * y**2 - z - 0.4 * np.cos(4 * np.pi * y) + 0.7)



# -*- coding: utf-8 -*-
import argparse
import importlib
import json
import logging
import os
import re
import sys
from io import StringIO

import boto3
import tabulate
import yaml
from dask.distributed import Client
from dask_kubernetes import KubeCluster
from kubernetes.client import Configuration
from kubernetes.client.api import core_v1_api
from kubernetes.config import load_kube_config

RUN_TEMPLATE = """
/bin/bash <<'EOF'

{}

EOF
"""

CONFIG_TEMPLATE = """
cat > config.json << JSON
{}
JSON
"""

WORKER_COMM = '/usr/bin/prepare.sh dask-worker --no-dashboard --memory-limit 0 --death-timeout 0'


def _import_function(config):
    function = config['function']
    function = function.split('.')
    function_name = function[-1]
    package = '.'.join(function[:-1])
    module = importlib.import_module(package)

    return getattr(module, function_name)


def _get_extra_setup(setup_dict):
    extra_packages = []

    script = setup_dict.get('script')
    if script:
        extra_packages.append('exec {}'.format(script))

    apt_packages = setup_dict.get('apt_packages')
    if apt_packages:
        extra_packages.append('apt get install {}'.format(' '.join(apt_packages)))

    pip_packages = setup_dict.get('pip_packages')
    if pip_packages:
        extra_packages.append('pip install {}'.format(' '.join(pip_packages)))

    git_repository = setup_dict.get('git_repository')
    if git_repository:
        url = git_repository.get('url')
        reference = git_repository.get('reference', 'master')
        install = git_repository.get('install')

        git_clone = 'git clone {} repo && cd repo'.format(url)
        git_checkout = 'git checkout {}'.format(reference)
        extra_packages.append('\n '.join([git_clone, git_checkout, install]))

    if len(extra_packages) > 1:
        return '\n '.join(extra_packages)

    return extra_packages[0]


def _generate_cluster_spec(config, kubernetes=False):
    extra_setup = ''
    dask_cluster = config['dask_cluster']
    metadata = {}

    worker_config = dask_cluster.get('worker_config')
    if worker_config.get('setup'):
        extra_setup = _get_extra_setup(worker_config['setup'])

    if kubernetes:
        name = worker_config.get('image', 'daskdev/dask:latest')
        name = '{}-'.format(re.sub(r'[\W_]', '-', name))
        metadata['generateName'] = name

        config_command = CONFIG_TEMPLATE.format(json.dumps(config))
        run_command = 'python -u -m btb_benchmark.kubernetes config.json'
        extra_setup = '\n'.join([extra_setup, config_command, run_command])

    else:
        run_command = WORKER_COMM
        extra_setup = '\n'.join([extra_setup, run_command])

    run_commands = RUN_TEMPLATE.format(extra_setup)

    spec = {
        'metadata': metadata,
        'spec': {
            'restartPolicy': 'Never',
            'containers': [{
                'args': ['-c', run_commands],
                'command': ['tini', '-g', '--', '/bin/sh'],
                'image': worker_config.get('image', 'daskdev/dask:latest'),
                'name': 'dask-worker',
                'resources': worker_config.get('resources', {})
            }]
        }
    }

    return spec


def _df_to_csv_str(df):
    with StringIO() as sio:
        df.to_csv(sio)
        return sio.getvalue()


def _upload_to_s3(bucket, path, results, aws_key=None, aws_secret=None):
    client = boto3.client('s3', aws_access_key_id=aws_key, aws_secret_access_key=aws_secret)
    client.put_object(Bucket=bucket, Key=path, Body=_df_to_csv_str(results))


def run_dask_function(config):
    """Start a Dask Cluster using dask-kubernetes and run a function.

    Talks to kubernetes to create `n` amount of new `pods` with a dask worker inside of each
    forming a `dask` cluster. Then, a function specified from `config` is being imported and
    run with the given arguments. The tasks created by this `function` are being run on the
    `dask` cluster for distributed computation.

    The config dict must contain the following sections:
        * run
        * dask_cluster
        * output

    Args:
        config (dict):
            Config dictionary.
    """
    output_conf = config.get('output')
    if output_conf:
        path = output_conf.get('path')
        if not path:
            raise ValueError('An output path must be provided when providing `output`.')

    cluster_spec = _generate_cluster_spec(config, kubernetes=False)
    cluster = KubeCluster.from_dict(cluster_spec)

    workers = config['dask_cluster'].get('workers')

    if not workers:
        cluster.adapt()
    elif isinstance(workers, int):
        cluster.scale(workers)
    else:
        cluster.adapt(**workers)

    client = Client(cluster)
    client.get_versions(check=True)

    try:
        run = _import_function(config['run'])
        kwargs = config['run']['args']
        results = run(**kwargs)

    finally:
        client.close()
        cluster.close()

    if output_conf:
        bucket = output_conf.get('bucket')

        try:
            if bucket:
                aws_key = output_conf.get('key')
                aws_secret = output_conf.get('secret_key')
                _upload_to_s3(bucket, path, results, aws_key, aws_secret)
            else:
                os.makedirs(os.path.dirname(path), exist_ok=True)
                results.to_csv(path)

        except Exception:
            print('Error storing results. Falling back to console dump.')
            print(_df_to_csv_str(results))

    else:
        return results


def run_on_kubernetes(config, namespace='default'):
    """Run dask function inside a pod using the given config.

     Create a pod, using the local kubernetes configuration that starts a Dask Cluster
     using dask-kubernetes and runs a function specified within the `config` dictionary.

    Args:
        config (dict):
            Config dictionary.
        namespace (str):
            Kubernetes namespace were the pod will be created.
    """
    # read local config
    load_kube_config()
    c = Configuration()
    Configuration.set_default(c)

    # create client and create pod on default namespace
    core_v1 = core_v1_api.CoreV1Api()
    spec = _generate_cluster_spec(config, kubernetes=True)
    core_v1.create_namespaced_pod(body=spec, namespace=namespace)
    print('Pod created.')


def _get_parser():
    parser = argparse.ArgumentParser(description='Run on Kubernetes Command Line Interface')

    parser.add_argument('config', help='Path to the JSON config file.')
    parser.add_argument('-v', '--verbose', action='count', default=0,
                        help='Be verbose. Use -vv for increased verbosity.')
    parser.add_argument('--create-pod', action='store_true',
                        help='Create a master pod and run the given `config` from there.')
    parser.add_argument('-n', '--namespace', default='default',
                        help='Namespace were the pod will be created.')

    return parser


def main():
    # Parse args
    parser = _get_parser()
    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(0)

    args = parser.parse_args()

    # Logger setup
    log_level = (3 - args.verbose) * 10
    fmt = '%(asctime)s - %(process)d - %(levelname)s - %(name)s - %(module)s - %(message)s'
    logging.basicConfig(level=log_level, format=fmt)

    with open(args.config) as config_file:
        if args.config.endswith('yaml') or args.config.endswith('yml'):
            config = yaml.safe_load(config_file)
        else:
            config = json.load(config_file)

    if args.create_pod:
        run_on_kubernetes(config, args.namespace)
    else:
        results = run_dask_function(config)

        if results is not None:
            print(tabulate.tabulate(
                results,
                tablefmt='github',
                headers=results.columns
            ))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
import logging
import random
import socket
from datetime import datetime, timedelta

import dask
import pandas as pd
from distributed.client import futures_of
from distributed.diagnostics.progressbar import TextProgressBar

from baytune.tuning.tuners.base import BaseTuner
from btb_benchmark.challenges import (
    MATH_CHALLENGES,
    RandomForestChallenge,
    SGDChallenge,
    XGBoostChallenge,
)
from btb_benchmark.challenges.challenge import Challenge
from btb_benchmark.challenges.datasets import get_dataset_names
from btb_benchmark.results import load_results, write_results
from btb_benchmark.tuning_functions import get_all_tuning_functions
from btb_benchmark.tuning_functions.btb import make_btb_tuning_function

LOGGER = logging.getLogger(__name__)
ALL_TYPES = ["math", "xgboost"]


def get_math_challenge_instance(name, *args, **kwargs):
    return MATH_CHALLENGES.get(name)()


CHALLENGE_GETTER = {
    "math": get_math_challenge_instance,
    "random_forest": RandomForestChallenge,
    "sgd": SGDChallenge,
    "xgboost": XGBoostChallenge,
}


@dask.delayed
def _evaluate_tuner_on_challenge(name, tuner, challenge, iterations):
    tunable_hyperparameters = challenge.get_tunable_hyperparameters()
    LOGGER.info("Evaluating tuner %s on challenge %s for %s iterations",
                name, challenge, iterations)
    try:
        start = datetime.utcnow()
        score = tuner(challenge.evaluate, tunable_hyperparameters, iterations)
        result = {
            "challenge": str(challenge),
            "tuner": name,
            "score": score,
            "iterations": iterations,
            "elapsed": datetime.utcnow() - start,
            "hostname": socket.gethostname()
        }
        if hasattr(challenge, "data"):
            result["rows"] = challenge.data[0].shape[0]

    except Exception as ex:
        LOGGER.warn(
            "Could not score tuner %s with challenge %s, error: %s", name, challenge, ex)
        result = {
            "challenge": str(challenge),
            "tuner": name,
            "score": None,
            "elapsed": datetime.utcnow() - start,
            "hostname": socket.gethostname()
        }

    return result


def _evaluate_tuners_on_challenge(tuners, challenge, iterations):
    LOGGER.info("Evaluating challenge %s", challenge)
    results = []
    for name, tuner in tuners.items():
        try:
            result = _evaluate_tuner_on_challenge(name, tuner, challenge, iterations)
            results.append(result)
        except Exception as ex:
            LOGGER.warn(
                "Could not score tuner %s with challenge %s, error: %s", name, challenge, ex)

    return results


class LogProgressBar(TextProgressBar):
    last = 0
    logger = logging.getLogger("distributed")

    def _draw_bar(self, remaining, all, **kwargs):
        frac = (1 - remaining / all) if all else 0

        if frac > self.last + 0.01:
            self.last = int(frac * 100) / 100
            bar = "#" * int(self.width * frac)
            percent = int(100 * frac)

            time_per_task = self.elapsed / (all - remaining)
            remaining_time = timedelta(seconds=time_per_task * remaining)
            eta = datetime.utcnow() + remaining_time

            elapsed = timedelta(seconds=self.elapsed)
            msg = "[{0:<{1}}] | {2}% Completed | {3} | {4} | {5}".format(
                bar, self.width, percent, elapsed, remaining_time, eta
            )
            self.logger.info(msg)

    def _draw_stop(self, **kwargs):
        pass


def progress(*futures):
    futures = futures_of(futures)
    if not isinstance(futures, (set, list)):
        futures = [futures]

    LogProgressBar(futures)


def benchmark(tuners, challenges, iterations, detailed_output=False):
    """Score ``tuners`` against a list of ``challenges`` for the given amount of iterations.

    This function scores a collection of ``tuners`` against a collection of ``challenges``
    performing tuning iterations in order to obtain a better score. At the end, the best score
    for each tuner / challenge is being returned. This data is returned as a ``pandas.DataFrame``.

    Args:
        tuners (dict):
            Python dictionary with the ``name`` of the function as ``key`` and the callable
            function that returns the best score for a given ``scorer``.
            This function must have three arguments:

                * scorer (function):
                    A function that performs scoring over params.
                * tunable (baytune.tuning.Tunable):
                    A ``Tunable`` instance used to instantiate a tuner.
                * iterations (int):
                    Number of tuning iterations to perform.

        challenges (list):
            A list of ``chalenges``. This challenges must inherit from
            ``baytune.challenges.challenge.Challenge``.
        iterations (int):
            Amount of tuning iterations to perform for each tuner and each challenge.
        detailed_output (bool):
            If ``True`` a dataframe with the elapsed time, score and iterations will be returned.

    Returns:
        pandas.DataFrame:
            A ``pandas.DataFrame`` with the obtained scores for the given challenges is being
            returned.
    """
    delayed = []

    for challenge in challenges:
        result = _evaluate_tuners_on_challenge(tuners, challenge, iterations)
        delayed.extend(result)

    persisted = dask.persist(*delayed)

    try:
        progress(persisted)
    except ValueError:
        # Using local client. No progress bar needed.
        pass

    results = dask.compute(*persisted)

    df = pd.DataFrame.from_records(results)
    if detailed_output:
        return df

    df = df.pivot(index="challenge", columns="tuner", values="score")
    df.index.rename(None, inplace=True)
    df.columns.rename(None, inplace=True)

    return df


def _as_list(param):
    """Make sure that param is either ``None`` or a ``list``."""
    if param is None or isinstance(param, (list, tuple)):
        return param

    return [param]


def _challenges_as_list(param):
    """Make sure that param is either ``None`` or a ``list``."""
    if param is None or isinstance(param, (list, tuple)):
        return param

    return get_dataset_names(param)


def _get_tuners_dict(tuners=None):
    all_tuners = get_all_tuning_functions()
    if tuners is None:
        LOGGER.info("Using all tuning functions.")
        return all_tuners
    else:
        selected_tuning_functions = {}
        for tuner in _as_list(tuners):
            if isinstance(tuner, type) and issubclass(tuner, BaseTuner):
                selected_tuning_functions[tuner.__name__] = make_btb_tuning_function(tuner)
            elif callable(tuner):
                selected_tuning_functions[tuner.__name__] = tuner
            else:
                tuning_function = all_tuners.get(tuner)
                if tuning_function:
                    LOGGER.info("Loading tuning function: %s", tuner)
                    selected_tuning_functions[tuner] = tuning_function
                else:
                    LOGGER.info("Could not load tuning function: %s", tuner)

        if not selected_tuning_functions:
            raise ValueError("No tunable function was loaded.")

        return selected_tuning_functions


def _get_all_challenge_names(challenge_types=None):
    all_challenge_names = []
    if "math" in challenge_types:
        all_challenge_names += list(MATH_CHALLENGES.keys())
    if any(name in challenge_types for name in ("sdg", "xgboost", "random_forest")):
        all_challenge_names += get_dataset_names("all")

    return all_challenge_names


def _get_challenges_list(challenges=None, challenge_types=None, sample=None, max_rows=None):
    challenge_types = _as_list(challenge_types) or ALL_TYPES
    challenges = _challenges_as_list(challenges) or _get_all_challenge_names(challenge_types)

    selected = []
    unknown = []

    if sample:
        if sample > len(challenges):
            raise ValueError("Sample can not be greater than {}".format(len(challenges)))

        challenges = random.sample(challenges, sample)

    for challenge in challenges:
        known = False
        if isinstance(challenge, Challenge):
            selected.append(challenge)
        else:
            for challenge_type in challenge_types:
                try:
                    challenge_class = CHALLENGE_GETTER[challenge_type]
                    challenge_instance = challenge_class(challenge, max_rows=max_rows)

                    if challenge_instance:
                        known = True
                        selected.append(challenge_instance)
                except Exception:
                    pass

            if not known:
                unknown.append(challenge)

    if unknown:
        raise ValueError("Challenges {} not of type {}".format(unknown, challenge_types))

    if not selected:
        raise ValueError("No challenges selected!")

    return selected


def run_benchmark(tuners=None, challenge_types=None, challenges=None, sample=None,
                  iterations=100, max_rows=None, output_path=None, detailed_output=False):
    """Execute the benchmark function and optionally store the result as a ``CSV``.

    This function provides a user-friendly interface to interact with the ``benchmark``
    function. It allows the user to specify an ``output_path`` where the results can be
    stored. If this path is not provided, a ``pandas.DataFrame`` will be returned.

    Args:
        tuners (str, baytune.tuning.tuners.base.BaseTuner or list):
            Tuner name, ``baytune.tuning.tuners.base.BaseTuner`` subclass or a list with the previously
            described objects. If ``None`` all available ``tuners`` implemented in
            ``btb_benchmark`` will be used.
        challenge_types (str or list):
            Type or list of types for challenges to be benchmarked, if ``None`` all available
            types will be used.
        challenges (str or list):
            If ``str`` it will be interpreted as ``collection`` of datasets (currently: all or
            openml100). A list containing: challenge name, ``btb_benchmark.challenge.Challenge``
            instance or a list with the previously described objects. If ``None`` will use
            ``challenge_types`` to determine which challenges to use.
        sample (int):
            Run only on a subset of the available datasets of the given size.
        iterations (int):
            Number of tuning iterations to perform per challenge and tuner.
        max_rows (int):
            Maximum number of rows to use from each dataset. If ``None``, or if the
            given number is higher than the number of rows in the dataset, the entire
            dataset is used. Defaults to ``None``.
        output_path (str):
            If an ``output_path`` is given, the final results will be saved in that location.
        detailed_output (bool):
            If ``True`` a dataframe with the elapsed time, score and iterations will be returned.

    Returns:
        pandas.DataFrame or None:
            If ``output_path`` is ``None`` it will return a ``pandas.DataFrame`` object,
            else it will dump the results in the specified ``output_path``.
    """
    tuners = _get_tuners_dict(tuners)
    challenges = _get_challenges_list(
        challenges=challenges,
        challenge_types=challenge_types,
        sample=sample,
        max_rows=max_rows
    )

    results = benchmark(tuners, challenges, iterations, detailed_output)

    if output_path:
        LOGGER.info("Saving benchmark report to %s", output_path)
        results.to_csv(output_path)

    else:
        return results


def summarize_results(input_paths, output_path):
    """Load multiple benchmark results CSV files and compile a summary.

    The result is an Excel file with one tab for each results CSV file
    and an additional Number of Wins tab with a summary of the number
    of challenges in which each Tuner got the best score.

    Args:
        inputs_paths (list[str]):
            List of paths to CSV files where the benchmarks results are stored.
            These files must have one column per Tuner and one row per Challenge.
        output_path (str):
            Path, including the filename, where the Excel file will be created.
    """
    results = load_results(input_paths)
    write_results(results, output_path)



import random

import numpy as np

from baytune.selection.selector import Selector


class UCB1(Selector):
    """UCB1 selector

    Uses Upper Confidence Bound 1 algorithm (UCB1) for bandit selection.

    See also::

       Auer, Peter et al. "Finite-time Analysis of the Multiarmed Bandit Problem."
       Machine Learning 47 (2002): 235-256.
    """

    def _shuffle(self, iterable):
        iterable = list(iterable)
        inds = list(range(len(iterable)))
        random.shuffle(inds)
        for i in inds:
            yield iterable[i]

    def bandit(self, choice_rewards):
        """
        Multi-armed bandit method which chooses the arm for which the upper
        confidence bound (UCB) of expected reward is greatest.

        If there are multiple arms with the same UCB1 index, then one is chosen
        at random.

        An explanation is here:
        https://www.cs.bham.ac.uk/internal/courses/robotics/lectures/ucb1.pdf
        """

        # count the larger of 1 and the total number of arm pulls
        total_pulls = max(1, sum(len(r) for r in choice_rewards.values()))

        def ucb1(choice):
            rewards = choice_rewards[choice]
            choice_pulls = max(len(rewards), 1)
            average_reward = np.nanmean(rewards) if len(rewards) else 0
            error = np.sqrt(2.0 * np.log(total_pulls) / choice_pulls)
            return average_reward + error

        return max(self._shuffle(choice_rewards), key=ucb1)



import logging

import numpy as np

from baytune.selection.ucb1 import UCB1

# the minimum number of scores that each choice must have in order to use best-K
# optimizations. If not all choices meet this threshold, default UCB1 selection
# will be used.
K_MIN = 2

logger = logging.getLogger("btb")


class BestKReward(UCB1):
    """Best K reward selector

    Computes the average reward from the past scores by using only the highest k scores. In
    implementation, the other scores are replaced with ``nan``s such that they still factor into
    the number of arm pulls.

    Args:
        k (int): number of best scores to consider
    """

    def __init__(self, choices, k=K_MIN):
        super(BestKReward, self).__init__(choices)
        self.k = k

    def compute_rewards(self, scores):
        """Retain the K best scores, and replace the rest with nans"""
        if len(scores) > self.k:
            scores = np.copy(scores)
            inds = np.argsort(scores)[: -self.k]
            scores[inds] = np.nan

        return list(scores)

    def select(self, choice_scores):
        """Select a choice using the K best scores

        Keeps the choice counts intact, but only let the bandit see the top k learners' scores.
        If there is not enough score history to do K-selection, use the default UCB1 reward
        function.
        """
        min_num_scores = min(len(s) for s in choice_scores.values())
        if min_num_scores >= K_MIN:
            logger.info(
                "{klass}: using Best K bandit selection".format(
                    klass=type(self).__name__
                )
            )
            compute_rewards = self.compute_rewards
        else:
            logger.warning(
                "{klass}: Not enough choices to do K-selection; using plain UCB1".format(
                    klass=type(self).__name__
                )
            )
            compute_rewards = super(BestKReward, self).compute_rewards

        # convert the raw scores list for each choice to a "rewards" list
        choice_rewards = {
            choice: compute_rewards(choice_scores[choice])
            for choice in choice_scores
            if choice in self.choices
        }

        return self.bandit(choice_rewards)


class BestKVelocity(BestKReward):
    """Best K velocity selector"""

    def compute_rewards(self, scores):
        """Compute the velocity of the best scores

        The velocities are the k distances between the k+1 best scores.
        """
        k = self.k
        m = max(len(scores) - k, 0)
        best_scores = sorted(scores)[-k - 1 :]
        velocities = np.diff(best_scores)
        nans = np.full(m, np.nan)
        return list(velocities) + list(nans)



import logging

from baytune.selection.ucb1 import UCB1

# the minimum number of scores that each choice must have in order to use
# best-K optimizations. If not all choices meet this threshold, default UCB1
# selection will be used.
K_MIN = 2

logger = logging.getLogger("btb")


class RecentKReward(UCB1):
    """Recent K reward selector

    Args:
        k (int): number of best scores to consider
    """

    def __init__(self, choices, k=K_MIN):
        super(RecentKReward, self).__init__(choices)
        self.k = k

    def compute_rewards(self, scores):
        """Retain the K most recent scores, and replace the rest with zeros"""
        for i in range(len(scores)):
            if i >= self.k:
                scores[i] = 0.0
        return scores

    def select(self, choice_scores):
        """Use the top k learner's scores for usage in rewards for the bandit calculation"""
        # if we don't have enough scores to do K-selection, fall back to UCB1
        min_num_scores = min([len(s) for s in choice_scores.values()])
        if min_num_scores >= K_MIN:
            logger.info(
                "{klass}: using Best K bandit selection".format(
                    klass=type(self).__name__
                )
            )
            reward_func = self.compute_rewards
        else:
            logger.warning(
                "{klass}: Not enough choices to do K-selection; using plain UCB1".format(
                    klass=type(self).__name__
                )
            )
            reward_func = super(RecentKReward, self).compute_rewards

        choice_rewards = {}
        for choice, scores in choice_scores.items():
            if choice not in self.choices:
                continue
            choice_rewards[choice] = reward_func(scores)

        return self.bandit(choice_rewards)


class RecentKVelocity(RecentKReward):
    """Recent K velocity selector"""

    def compute_rewards(self, scores):
        """Compute the velocity of thte k+1 most recent scores.

        The velocity is the average distance between scores. Return a list with those k velocities
        padded out with zeros so that the count remains the same.
        """
        # take the k + 1 most recent scores so we can get k velocities
        recent_scores = scores[: -self.k - 2 : -1]
        velocities = [
            recent_scores[i] - recent_scores[i + 1]
            for i in range(len(recent_scores) - 1)
        ]
        # pad the list out with zeros, so the length of the list is
        # maintained
        zeros = (len(scores) - self.k) * [0]
        return velocities + zeros



# -*- coding: utf-8 -*-

from skopt import gp_minimize, space
from skopt.utils import use_named_args


def _dimension_space_from_dict(dict_hyperparams):
    hyperparams = list()

    if not isinstance(dict_hyperparams, dict):
        raise TypeError('Hyperparams must be a dictionary.')

    for name, hyperparam in dict_hyperparams.items():
        hp_type = hyperparam['type']

        if hp_type == 'int':
            hp_range = hyperparam.get('range') or hyperparam.get('values')
            hp_min = min(hp_range) if hp_range else None
            hp_max = max(hp_range) if hp_range else None
            hp_instance = space.Integer(hp_min, hp_max, name=name)

        elif hp_type == 'float':
            hp_range = hyperparam.get('range') or hyperparam.get('values')
            hp_min = min(hp_range)
            hp_max = max(hp_range)
            hp_instance = space.Real(hp_min, hp_max, name=name)

        elif hp_type == 'bool':
            hp_instance = space.Categorical([True, False], name=name)

        elif hp_type == 'str':
            hp_choices = hyperparam.get('range') or hyperparam.get('values')
            hp_instance = space.Categorical(hp_choices, name=name)

        hyperparams.append(hp_instance)

    return hyperparams


def _make_minimize_function(scoring_function, space):
    """Create a minimize function.

    Given a maximization ``scoring_function`` convert it to minimize in order to work with
    ``hyperopt``, as ``benchmark`` works with ``maximization``.

    Also ``hyperopt`` params are being passed as a python ``dict``, we pass those as ``kwargs``
    to the ``scoring_function``.
    """
    @use_named_args(space)
    def minimized_function(**params):
        return -scoring_function(**params)

    return minimized_function


def _skopt_tuning_function(scoring_function, tunable_hyperparameters, iterations, acq_function):

    space = _dimension_space_from_dict(tunable_hyperparameters)
    scoring_function = _make_minimize_function(scoring_function, space)

    res = gp_minimize(
        scoring_function,
        space,
        acq_func=acq_function,
        n_calls=iterations,
        random_state=0
    )
    return -res.fun


def skopt_LCB(scoring_function, tunable_hyperparameters, iterations):
    """Lower Confidence Bound acquisition function."""
    return _skopt_tuning_function(
        scoring_function,
        tunable_hyperparameters,
        iterations,
        'LCB',
    )


def skopt_EI(scoring_function, tunable_hyperparameters, iterations):
    """Expected Improvement acquisition function."""
    return _skopt_tuning_function(
        scoring_function,
        tunable_hyperparameters,
        iterations,
        'EI',
    )


def skopt_PI(scoring_function, tunable_hyperparameters, iterations):
    """Probability of Improvement acquisition function."""
    return _skopt_tuning_function(
        scoring_function,
        tunable_hyperparameters,
        iterations,
        'PI',
    )


def skopt_gp_hedge(scoring_function, tunable_hyperparameters, iterations):
    """Probabilistically choose between `LCB`, `EI` or `PI` acquisition function while tuning."""
    return _skopt_tuning_function(
        scoring_function,
        tunable_hyperparameters,
        iterations,
        'gp_hedge',
    )



# -*- coding: utf-8 -*-

import itertools
import json
import logging
from collections import Counter, defaultdict
from hashlib import md5

import numpy as np
from tqdm.auto import trange

from baytune.selection.ucb1 import UCB1
from baytune.tuning.tunable import Tunable
from baytune.tuning.tuners.base import StopTuning
from baytune.tuning.tuners.gaussian_process import GPTuner

LOGGER = logging.getLogger(__name__)


class BTBSession:
    """BTBSession class.

    A ``BTBSession`` represents the process of selecting and tuning several tunables
    until the best possible configuration for a specific ``scorer`` is found.

    For this, a loop is run in which for each iteration a combination of a ``Selector`` and
    ``Tuner`` is used to decide which tunable to score next and with which hyperparameters.

    While running, the ``BTBSession`` handles the errors discarding, if configured to do so,
    the tunables that have reached as many errors as the user specified.

    Attributes:
        best_proposal (dict):
            Best configuration found with the name of the tunable and the hyperparameters
            and crossvalidated score obtained for it.
        best_score (float):
            Best score obtained for this session so far.
        proposals (dict):
            Dictionary containing all the proposals generated by the ``BTBSession``.
        iterations (int):
            Amount of iterations run.
        errors (Counter):
            A Counter of the errors that each Tunable had during the session.

    Args:
        tunables (dict):
            Python dictionary that has as keys the name of the tunable and
            as value a dictionary with the tunable hyperparameters or an
            ``baytune.tuning.tunable.Tunable`` instance.
        scorer (callable object / function):
            A callable object or function with signature ``scorer(tunable_name, config)``
            wich should return only a single value.
        tuner_class (baytune.tuning.tuner.BaseTuner):
            A tuner based on BTB ``BaseTuner`` class. This tuner will manage the new proposals.
            Defaults to ``baytune.tuning.tuners.gaussian_process.GPTuner``
        selector_class (baytune.selection.selector.Selector):
            A selector based on BTB ``Selector`` class. This will determinate which one of
            the tunables is performing better, and which one to test next. Defaults to
            ``baytune.selection.selectors.ucb1.UCB1``
        maximize (bool):
            If ``True`` the scores are interpreted as bigger is better, if ``False`` then smaller
            is better, this should depend on the problem type (maximization or minimization).
            Defaults to ``True``.
        max_erors (int):
            Amount of errors allowed for a tunable to not generate a score. Once this amount of
            errors is reached, the tunable will be removed from the list. Defaults to 1.
        verbose (bool):
            If ``True`` a progress bar will be displayed for the ``run`` process.
    """

    _tunables = None
    _scorer = None
    _tuner_class = None
    _selector = None
    _maximize = None
    _max_errors = None
    _best_normalized = None
    _tunable_names = None
    _normalized_scores = None
    _tuners = None
    _range = None

    best_proposal = None
    best_score = None
    proposals = None
    iterations = None
    errors = None

    def _normalize(self, score):
        if score is not None:
            return score if self._maximize else -score

    def __init__(
        self,
        tunables,
        scorer,
        tuner_class=GPTuner,
        selector_class=UCB1,
        maximize=True,
        max_errors=1,
        verbose=False,
    ):
        self._tunables = tunables
        self._scorer = scorer
        self._tuner_class = tuner_class
        self._tunable_names = list(self._tunables.keys())
        self._selector = selector_class(self._tunable_names)
        self._maximize = maximize
        self._max_errors = max_errors

        self.best_proposal = None
        self.proposals = dict()
        self.iterations = 0
        self.errors = Counter()
        self.best_score = None

        self._best_normalized = -np.inf
        self._normalized_scores = defaultdict(list)
        self._tuners = dict()
        self._range = trange if verbose else range

    def _make_dumpable(self, to_dump):
        dumpable = {}
        for key, value in to_dump.items():
            if not isinstance(key, str):
                key = str(key)

            if isinstance(value, np.integer):
                value = int(value)
            elif isinstance(value, np.floating):
                value = float(value)
            elif isinstance(value, np.ndarray):
                value = value.tolist()
            elif isinstance(value, np.bool_):
                value = bool(value)
            elif value == "None":
                value = None

            dumpable[key] = value

        return dumpable

    def _make_id(self, name, config):
        dumpable_config = self._make_dumpable(config)
        proposal = {
            "name": name,
            "config": dumpable_config,
        }
        hashable = json.dumps(proposal, sort_keys=True).encode()

        return md5(hashable).hexdigest()

    def _remove_tunable(self, tunable_name):
        """Remove a tunable from the candidates list.

        This is necessary when:
            - Duplicates are not allowed and the tunable has exhausted all its
              configurations.
            - The tunable has failed more than ``max_errors`` times.

        When this happens, the tunable is removved from the tunables dict
        and its scores are removed from the normmalized_scores dict used by
        the selectors.
        """
        self._normalized_scores.pop(tunable_name, None)
        self._tunables.pop(tunable_name, None)

    def _get_next_tunable_name(self):
        if self._normalized_scores:
            tunable_name = self._selector.select(self._normalized_scores)
        else:
            # if _normalized_scores is still empty the selector crashes
            # this happens when max_errors > 1, all tunables have tuners
            # and all previous trials have crashed.
            tunable_name = np.random.choice(list(self._tunables.keys()))

        return tunable_name

    def propose(self):
        """Propose a new configuration to score.

        Every time ``propose`` is called, a new tunable will be selected and a new
        hyperparameter proposal will be generated for it.

        At the begining, the default hyperparameters of each one of the tunables
        will be returned sequencially in the same order as they were passed to
        the ``BTBSession``.

        After that, once each tunable has been scored at least once, the tunable
        used to generate the new proposals will be selected optimally each time
        by the selector.

        If a tunable runs out of proposals, it will be discarded from the list and will
        not be proposed again.

        Finally, when all the tunables have ran out of proposals, a ``StopTuning`` exception
        will be raised.

        Returns:
            tuple (str, dict):
                * Name of the tunable to try next.
                * Hyperparameters proposal.

        Raises:
            StopTuning:
                If the ``BTBSession`` has run out of proposals to generate.
        """
        if not self._tunables:
            raise StopTuning("There are no tunables left to try.")

        if len(self._tuners) < len(self._tunable_names):
            tunable_name = self._tunable_names[len(self._tuners)]
            tunable = self._tunables[tunable_name]

            if isinstance(tunable, dict):
                LOGGER.info("Creating Tunable instance from dict.")
                tunable = Tunable.from_dict(tunable)

            if not isinstance(tunable, Tunable):
                raise TypeError(
                    "Tunable can only be an instance of baytune.tuning.Tunable or dict"
                )

            LOGGER.info("Obtaining default configuration for %s", tunable_name)
            config = tunable.get_defaults()

            if tunable.cardinality == 1:
                LOGGER.warn(
                    "Skipping tuner creation for Tunable %s with cardinality 1",
                    tunable_name,
                )
                tuner = None
            else:
                tuner = self._tuner_class(tunable)

            self._tuners[tunable_name] = tuner

        else:
            tunable_name = self._get_next_tunable_name()
            tuner = self._tuners[tunable_name]

            try:
                if tuner is None:
                    raise StopTuning(
                        "Tunable %s has no tunable hyperparameters", tunable_name
                    )

                LOGGER.info(
                    "Generating new proposal configuration for %s", tunable_name
                )
                config = tuner.propose(1)

            except StopTuning:
                LOGGER.info("%s has no more configs to propose.", tunable_name)
                self._remove_tunable(tunable_name)
                tunable_name, config = self.propose()

        proposal_id = self._make_id(tunable_name, config)
        self.proposals[proposal_id] = {
            "id": proposal_id,
            "name": tunable_name,
            "config": config,
        }

        return tunable_name, config

    def handle_error(self, tunable_name):
        """Handle errors when ``score`` is ``None``.

        If the given ``tunable_name`` accumulates more errors than ``self._max_errors``
        this is removed from the selector's choices.

        Args:
            tunable_name (str):
                The name of the tunable to which this configuration belongs.
        """
        self.errors[tunable_name] += 1
        errors = self.errors[tunable_name]

        if errors >= self._max_errors:
            LOGGER.warning(
                "Too many errors: %s. Removing tunable %s", errors, tunable_name
            )
            self._remove_tunable(tunable_name)

    def record(self, tunable_name, config, score):
        """Record the configuration and the obtained score to the tuner.

        If the score is the best one so far, the ``best_proposal`` and ``best_score`` are
        updated.

        Args:
            tunable_name (str):
                The name of the tunable to which this configuration belongs.
            config (dict):
                Hyperparameter proposal, as given by the tunable.
            score (float):
                Obtained score with the given configuration.
        """
        proposal_id = self._make_id(tunable_name, config)
        proposal = self.proposals[proposal_id]
        proposal["score"] = score

        if score is None:
            self.handle_error(tunable_name)
        else:
            normalized = self._normalize(score)
            self._normalized_scores[tunable_name].append(normalized)

            if normalized > self._best_normalized:
                LOGGER.info("New optimal found: %s - %s", tunable_name, score)
                self.best_proposal = proposal
                self.best_score = score
                self._best_normalized = normalized
            try:
                tuner = self._tuners[tunable_name]
                if tuner is None:
                    LOGGER.warn(
                        "Skipping record for Tunable %s with cardinality 1",
                        tunable_name,
                    )
                else:
                    tuner.record(config, normalized)

            except Exception:
                LOGGER.exception(
                    "Could not record configuration and score for tuner %s.",
                    tunable_name,
                )

    def run(self, iterations=None):
        """Run the selection and tuning loop for the given number of iterations.

        At each iteration, the `BTBSession` will generate a new proposal calling
        ``self.propose``, score it using the `self.scorer`, and finally record the
        obtained score back to the tuner calling `self.record`.

        If no iterations are given, run infinitely until interrupted or until all the
        tuner proposals are exhausted.

        Scoring errors will also be captured and recorded.

        Returns:
            best_proposal (dict):
                Best configuration found with the name of the tunable and the hyperparameters
                and crossvalidated score obtained for it.
        """
        if iterations is None:
            iterator = itertools.count()
        else:
            iterator = self._range(iterations)

        for _ in iterator:
            self.iterations += 1
            tunable_name, config = self.propose()

            try:
                LOGGER.debug(
                    "Scoring proposal %s - %s: %s",
                    self.iterations,
                    tunable_name,
                    config,
                )
                score = self._scorer(tunable_name, config)

            except Exception:
                params = "\n".join("{}: {}".format(k, v) for k, v in config.items())
                LOGGER.exception(
                    "Proposal %s - %s crashed with the following configuration: %s",
                    self.iterations,
                    tunable_name,
                    params,
                )

                score = None

            self.record(tunable_name, config, score)

        return self.best_proposal



# -*- coding: utf-8 -*-

"""Top level where all the acquisition functions are imported."""

from baytune.tuning.acquisition.expected_improvement import (
    ExpectedImprovementAcquisition,
)
from baytune.tuning.acquisition.predicted_score import PredictedScoreAcquisition

__all__ = ("ExpectedImprovementAcquisition", "PredictedScoreAcquisition")



import random

from baytune.selection.selector import Selector


class Uniform(Selector):
    """Uniform selector

    Selects a choice uniformly at random.
    """

    def select(self, choice_scores):
        return random.choice(self.choices)



# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod

import numpy as np


class BaseAcquisition(metaclass=ABCMeta):
    def __init_acquisition__(self, **kwargs):
        pass

    @staticmethod
    def _get_max_candidates(candidates, n):
        k = min(n, len(candidates) - 1)  # kth element
        sorted_candidates = np.argpartition(-candidates, k)

        return sorted_candidates[:n]

    @abstractmethod
    def _acquire(self, candidates, num_candidates=1):
        """Decide which candidates to return as proposals.

        Apply a decision function to select the best candidates from
        the predicted scores list.

        Once the best candidates are found, their indexes are returned.

        Args:
            candidates (numpy.ndarray):
                2D array with two columns: scores and standard deviations
            num_candidates (int):
                Number of candidates to return.

        Returns:
            numpy.ndarray:
                Selected candidates indexes.
        """
        pass



# -*- coding: utf-8 -*-

"""Top level of the tuning module."""

from baytune.tuning.hyperparams.boolean import BooleanHyperParam
from baytune.tuning.hyperparams.categorical import CategoricalHyperParam
from baytune.tuning.hyperparams.numerical import FloatHyperParam, IntHyperParam
from baytune.tuning.tunable import Tunable
from baytune.tuning.tuners.base import StopTuning
from baytune.tuning.tuners.gaussian_process import (
    GCPEiTuner,
    GCPTuner,
    GPEiTuner,
    GPTuner,
)
from baytune.tuning.tuners.uniform import UniformTuner

__all__ = (
    "BooleanHyperParam",
    "CategoricalHyperParam",
    "GCPEiTuner",
    "GCPTuner",
    "GPEiTuner",
    "GPTuner",
    "FloatHyperParam",
    "IntHyperParam",
    "StopTuning",
    "Tunable",
    "UniformTuner",
)



import logging

from sklearn.linear_model import SGDClassifier
from sklearn.metrics import f1_score

from btb_benchmark.challenges.mlchallenge import MLChallenge

LOGGER = logging.getLogger(__name__)


class SGDChallenge(MLChallenge):

    # TARGET
    TARGET_COLUMN = 'class'
    MAKE_BINARY = False

    # CROSS VALIDATE / SCORER
    METRIC = f1_score
    METRIC_ARGS = {'average': 'macro'}
    ENCODE = True
    STRATIFIED = True

    # MODEL
    MODEL_DEFAULTS = {'random_state': 0}
    MODEL = SGDClassifier
    TUNABLE_HYPERPARAMETERS = {
        "loss": {
            "type": "str",
            "default": "hinge",
            "range": [
                "log",
                "hinge",
                "modified_huber",
                "squared_hinge",
                "perceptron",
                "squared_loss",
                "huber",
                "epsilon_insensitive",
                "squared_epsilon_insensitive"
            ]
        },
        "penalty": {
            "type": "str",
            "default": None,
            "values": [
                None,
                "l2",
                "l1",
                "elasticnet"
            ]
        },
        "alpha": {
            "type": "float",
            "default": 0.0001,
            "values": [
                0.0001,
                1
            ]
        },
        "max_iter": {
            "type": "int",
            "default": 1000,
            "values": [
                1,
                5000
            ]
        },
        "tol": {
            "type": "float",
            "default": 1e-3,
            "values": [
                1e-3,
                1
            ]
        },
        "shuffle": {
            "type": "bool",
            "default": True,
        },
    }



# -*- coding: utf-8 -*-

"""Package where the tuners based on GaussianProcessMetaModel are defined."""

from baytune.tuning.acquisition.expected_improvement import (
    ExpectedImprovementAcquisition,
)
from baytune.tuning.acquisition.predicted_score import PredictedScoreAcquisition
from baytune.tuning.metamodels.gaussian_process import (
    GaussianCopulaProcessMetaModel,
    GaussianProcessMetaModel,
)
from baytune.tuning.tuners.base import BaseMetaModelTuner


class GPTuner(GaussianProcessMetaModel, PredictedScoreAcquisition, BaseMetaModelTuner):
    """Gaussian Process Tuner.

    This class uses a ``GaussianProcessRegressor`` model from the ``sklearn.gaussian_process``
    package, using a ``numpy.argmax`` function to return the better configurations predicted
    from the model.
    """

    def __init__(
        self,
        tunable,
        maximize=True,
        num_candidates=1000,
        min_trials=5,
        length_scale=0.1,
    ):
        """Create an instance of ``GPTuner``.

        Args:
            tunable (baytune.tuning.tunable.Tunable):
                Instance of a tunable class containing hyperparameters to be tuned.
            num_candidates (int):
                Number of samples to generate and select the best of it for each proposal.
                Defaults to 1000.
            maximize (bool):
                If ``True`` the model will understand that the score bigger is better, if ``False``
                the smaller is better. Defaults to ``True``.
            min_trials (int):
                Number of recorded ``trials`` needed to perform a fitting over the model.
                Defaults to 2.
            length_scale (float or array):
                A float or array with shape ``(n_features,)``, used for the default ``RBF`` kernel.
        """
        self._metamodel_kwargs = {"length_scale": length_scale}
        super().__init__(tunable, maximize, num_candidates, min_trials)

    def __repr__(self):
        length_scale = self._metamodel_kwargs.get("length_scale")
        args = (
            self.tunable,
            self.maximize,
            self.num_candidates,
            self.min_trials,
            length_scale,
        )
        return (
            "GPTuner(tunable={}, maximize={}, "
            "num_candidates={}, min_trials={}, "
            "length_scale={})"
        ).format(*args)


class GPEiTuner(
    GaussianProcessMetaModel, ExpectedImprovementAcquisition, BaseMetaModelTuner
):
    """Gaussian Process Expected Improvement Tuner.

    This class uses a ``GaussianProcessRegressor`` model from the ``sklearn.gaussian_process``
    package, using an ``ExpectedImprovement`` function to return the better configurations
    predicted from the model.
    """

    def __init__(
        self,
        tunable,
        maximize=True,
        num_candidates=1000,
        min_trials=5,
        length_scale=0.1,
    ):
        """Create an instance of ``GPEiTuner``.

        Args:
            tunable (baytune.tuning.tunable.Tunable):
                Instance of a tunable class containing hyperparameters to be tuned.
            num_candidates (int):
                Number of samples to generate and select the best of it for each proposal.
                Defaults to 1000.
            maximize (bool):
                If ``True`` the model will understand that the score bigger is better, if ``False``
                the smaller is better. Defaults to ``True``.
            min_trials (int):
                Number of recorded ``trials`` needed to perform a fitting over the model.
                Defaults to 2.
            length_scale (float or array):
                A float or array with shape ``(n_features,)``, used for the default ``RBF`` kernel.
        """
        self.length_scale = length_scale
        self._metamodel_kwargs = {"length_scale": self.length_scale}
        super().__init__(tunable, maximize, num_candidates, min_trials)

    def __repr__(self):
        length_scale = self._metamodel_kwargs.get("length_scale")
        args = (
            self.tunable,
            self.maximize,
            self.num_candidates,
            self.min_trials,
            length_scale,
        )
        return (
            "GPEiTuner(tunable={}, maximize={}, "
            "num_candidates={}, min_trials={}, "
            "length_scale={})"
        ).format(*args)


class GCPTuner(GaussianCopulaProcessMetaModel, GPTuner):
    """Gaussian Copula Process Tuner.

    This class uses a ``GaussianProcessRegressor`` model from the ``sklearn.gaussian_process``
    package, using a ``numpy.argmax`` function to return the better configurations predicted
    from the meta model that converts the input data using a ``Univariate`` copula.
    """

    def __repr__(self):
        length_scale = self._metamodel_kwargs.get("length_scale")
        args = (
            self.tunable,
            self.maximize,
            self.num_candidates,
            self.min_trials,
            length_scale,
        )
        return (
            "GCPTuner(tunable={}, maximize={}, "
            "num_candidates={}, min_trials={}, "
            "length_scale={})"
        ).format(*args)


class GCPEiTuner(GaussianCopulaProcessMetaModel, GPEiTuner):
    """Gaussian Copula Process Expected Improvement Tuner.

    This class uses a ``GaussianProcessRegressor`` model from the ``sklearn.gaussian_process``
    package, using an ``ExpectedImprovement`` function to return the better configurations
    predicted from the meta model that converts the input data using a ``Univariate`` copula.
    """

    def __repr__(self):
        length_scale = self._metamodel_kwargs.get("length_scale")
        args = (
            self.tunable,
            self.maximize,
            self.num_candidates,
            self.min_trials,
            length_scale,
        )
        return (
            "GCPEiTuner(tunable={}, maximize={}, "
            "num_candidates={}, min_trials={}, "
            "length_scale={})"
        ).format(*args)



# -*- coding: utf-8 -*-

"""Top level where all the metamodels are imported."""

from baytune.tuning.metamodels.gaussian_process import GaussianProcessMetaModel

__all__ = ("GaussianProcessMetaModel",)



# -*- coding: utf-8 -*-

from abc import ABCMeta
from copy import deepcopy

import numpy as np


class BaseMetaModel(metaclass=ABCMeta):
    """BaseMetaModel class.

    BaseMetaModel class is an abstract representation of a ``MetaModel`` that after being
    fitted can predict the score that is being expected to be obtained with the hyperparameter
    configuration proposed.

    Attributes:
        _MODEL_CLASS (type):
            Class to be instantiated and to be fited and used to generate predictions. This
            attribute must be overriden by the child class in order to create an actual model.
        _MODEL_KWARGS (dict):
            Dictionary with the default ``kwargs`` to be given for the ``self._MODEL_CLASS``.
        _model_kwargs (dict):
            A dictionary that is used during instantiation of the ``BaseMetaModelTuner`` in
            order to be able to give or set arguments for the model.
        _model (object):
            Instance of ``self._MODEL_CLASS``, defaults to ``None``.
    """

    _MODEL_CLASS = None
    _MODEL_KWARGS_DEFAULT = None
    _model_kwargs = None
    _model_instance = None

    def __init_metamodel__(self, **kwargs):
        pass

    def _init_model(self):
        """Create an instance of a ``self._MODEL_CLASS``.

        Generate ``self._model_instance`` by using the corresponding ``kwargs`` for
        ``self._MODEL_CLASS`` provided by the user and ``self._MODEL_KWARGS_DEFAULT``.
        """
        if self._MODEL_KWARGS_DEFAULT is not None:
            model_kwargs = deepcopy(self._MODEL_KWARGS_DEFAULT)

        else:
            model_kwargs = {}

        if self._model_kwargs:
            model_kwargs.update(self._model_kwargs)

        self._model_instance = self._MODEL_CLASS(**model_kwargs)

    def _fit(self, trials, scores):
        """Fit the internal meta-model.

        Create a new instance of ``self._META_CLASS`` and fit it over ``trials`` and ``scores``.

        Args:
            params (array-like):
                2D array-like with shape ``(len(trials), n_params)``.
            scores (array-like):
                Array-like with shape ``(len(trials), 1)``.
        """
        self._init_model()
        self._model_instance.fit(trials, scores)

    def _predict(self, candidates):
        """Predict performance for given candidates.

        Predict the performance for the given candidates using the ``self._META_CLASS`` instance
        and it's method ``predict``. Depending on the meta-model, the predictions could be point
        predictions or could also include a standard deviation at that point (like with a
        Gaussian Process meta-model).

        Args:
            candidates (array-like):
                2D array-like with shape ``(num_cadidates, num_params)``.
        Returns:
            predictions (array-like):
                2D array-like with shape ``(num_candidates, num_outputs)``.
        """
        return np.array(self._model_instance.predict(candidates))



# -*- coding: utf-8 -*-

from hyperopt import Trials, fmin, hp, tpe


def _search_space_from_dict(dict_hyperparams):
    hyperparams = {}

    if not isinstance(dict_hyperparams, dict):
        raise TypeError('Hyperparams must be a dictionary.')

    for name, hyperparam in dict_hyperparams.items():
        hp_type = hyperparam['type']

        if hp_type == 'int':
            hp_range = hyperparam.get('range') or hyperparam.get('values')
            hp_min = min(hp_range) if hp_range else None
            hp_max = max(hp_range) if hp_range else None
            hp_instance = hp.uniformint(name, hp_min, hp_max)

        elif hp_type == 'float':
            hp_range = hyperparam.get('range') or hyperparam.get('values')
            hp_min = min(hp_range)
            hp_max = max(hp_range)
            hp_instance = hp.uniform(name, hp_min, hp_max)

        elif hp_type == 'bool':
            hp_instance = hp.choice(name, [True, False])

        elif hp_type == 'str':
            hp_choices = hyperparam.get('range') or hyperparam.get('values')
            hp_instance = hp.choice(name, hp_choices)

        hyperparams[name] = hp_instance

    return hyperparams


def _make_minimize_function(scoring_function):
    """Create a minimize function.

    Given a maximization ``scoring_function`` convert it to minimize in order to work with
    ``hyperopt``, as ``benchmark`` works with ``maximization``.

    Also ``hyperopt`` params are being passed as a python ``dict``, we pass those as ``kwargs``
    to the ``scoring_function``.
    """
    def minimized_function(params):
        return -scoring_function(**params)

    return minimized_function


def _hyperopt_tuning_function(algo, scoring_function, tunable_hyperparameters, iterations):
    """Create a tuning function that uses ``HyperOpt``.

    With a given suggesting algorithm from the library ``HyperOpt``, create a tuning
    function that maximize the score, using ``fmin``.

    Args:
        algo (hyperopt.algo):
            Search / Suggest ``HyperOpt`` algorithm to be used with ``fmin`` function.
    """

    minimized_scoring = _make_minimize_function(scoring_function)
    search_space = _search_space_from_dict(tunable_hyperparameters)
    trials = Trials()
    fmin(
        minimized_scoring,
        search_space,
        algo=algo,
        max_evals=iterations,
        trials=trials,
        verbose=False
    )

    # normalize best score to match other tuners
    best_score = -1 * trials.best_trial['result']['loss']

    return best_score


def hyperopt_tpe(scoring_function, tunable_hyperparameters, iterations):
    """Tree-structured Parzen Estimator"""
    return _hyperopt_tuning_function(
        tpe.suggest,
        scoring_function,
        tunable_hyperparameters,
        iterations
    )



# -*- coding: utf-8 -*-
"""SMAC: Sequential Model-based Algorithm Configuration"""

from tempfile import TemporaryDirectory

import ConfigSpace.hyperparameters as hp
from smac.configspace import ConfigurationSpace
from smac.facade.hyperband_facade import HB4AC
from smac.facade.smac_hpo_facade import SMAC4HPO
from smac.optimizer import acquisition
from smac.scenario.scenario import Scenario

_NONE = '__NONE__'


def _create_config_space(dict_hyperparams):
    """Create the hyperparameters hyperspace."""
    config_space = ConfigurationSpace()

    if not isinstance(dict_hyperparams, dict):
        raise TypeError('Hyperparams must be a dictionary.')

    for name, hyperparam in dict_hyperparams.items():
        hp_type = hyperparam['type']

        if hp_type == 'int':
            hp_range = hyperparam.get('range') or hyperparam.get('values')
            hp_min = min(hp_range)
            hp_max = max(hp_range)
            hp_default = hyperparam.get('default') or hp_min
            config_space.add_hyperparameter(
                hp.UniformIntegerHyperparameter(name, hp_min, hp_max, default_value=hp_default))

        elif hp_type == 'float':
            hp_range = hyperparam.get('range') or hyperparam.get('values')
            hp_min = min(hp_range)
            hp_max = max(hp_range)
            hp_default = hyperparam.get('default') or hp_min
            config_space.add_hyperparameter(
                hp.UniformFloatHyperparameter(name, hp_min, hp_max, default_value=hp_default))

        elif hp_type == 'bool':
            hp_default = bool(hyperparam.get('default'))
            config_space.add_hyperparameter(
                hp.CategoricalHyperparameter(name, ['true', 'false'], default_value=hp_default))

        elif hp_type == 'str':
            hp_range = hyperparam.get('range') or hyperparam.get('values')
            hp_range = [_NONE if hp is None else hp for hp in hp_range]
            hp_default = hyperparam.get('default') or hp_range[0]
            hp_default = _NONE if hp_default is None else hp_default

            config_space.add_hyperparameter(
                hp.CategoricalHyperparameter(name, hp_range, default_value=hp_default))

    return config_space


def _parse_params(params):
    parsed_params = dict()
    params = params if isinstance(params, dict) else params.get_dictionary()

    for key, value in params.items():
        parsed_params[key] = None if value == _NONE else value

    return parsed_params


def _adapt_scoring_function(scoring_function):
    """Adapt the scoring function.

    SMAC optimize function calls the scoring function with a dict object, however our challenges
    recieve them as kwargs. The optimize function is ment to minimize, thats why we also return
    a negative score.
    """

    def adapted_function(params):
        parsed_params = _parse_params(params)
        return -scoring_function(**parsed_params)

    return adapted_function


def _get_optimizer_params(scoring_function, tunable_hyperparameters,
                          iterations, tmp_dir, **kwargs):
    config_space = _create_config_space(tunable_hyperparameters)
    tae_runner = _adapt_scoring_function(scoring_function)
    scenario = Scenario({
        'run_obj': 'quality',
        'runcount_limit': iterations,
        'cs': config_space,
        'deterministic': 'true',
        'output_dir': tmp_dir,
        'limit_resources': False,
    })

    optimizer_params = {
        'scenario': scenario,
        'rng': 42,
        'tae_runner': tae_runner,
    }

    if kwargs:
        optimizer_params.update(kwargs)

    return optimizer_params


def _smac_tuning_function(optimizer, scoring_function,
                          tunable_hyperparameters, iterations, **kwargs):
    """Construct and run a full optimization loop.

    Given an optimizer from ``smac.facade``, use it to perform a complete optimization for
    a given ``scoring_function``. This is achieved by creating a ``config_space`` from the
    tunable hyperparameters, adapting the scoring function to work with minimization, and
    then, create an instace scenario that with the config space and the amount of iterations.

    Finally we use the optimizer with the previously created sceneario and adapted scoring
    function to optimize this and obtain the best configuration for the given iterations.

    Args:
        optimizer (type):
            A ``smac.facade`` class that represents a tuner.
        scoring_function (function):
            A function that performs scoring over params.
        tunable_hyperparameters (dict):
            A python dict with hyperparameters.
        iterations (int):
            Number of tuning iterations to perform.
        kwargs (kwargs):
            Any additional configuration used by the optimizer can be passed as
            keyword args.
    """
    with TemporaryDirectory() as tmp_dir:
        optimizer_params = _get_optimizer_params(
            scoring_function,
            tunable_hyperparameters,
            iterations,
            tmp_dir,
            **kwargs
        )

        smac = optimizer(**optimizer_params)
        best_config = smac.optimize()

    return scoring_function(**_parse_params(best_config))


def smac_smac4hpo_ei(scoring_function, tunable_hyperparameters, iterations):
    """Use SMAC4HPO with Expected Improvement acquisition function."""
    return _smac_tuning_function(
        SMAC4HPO,
        scoring_function,
        tunable_hyperparameters,
        iterations,
        acquisition_function=acquisition.EI
    )


def smac_smac4hpo_lcb(scoring_function, tunable_hyperparameters, iterations):
    """Use SMAC4HPO with Lower Confidence Bound acquisition function."""
    return _smac_tuning_function(
        SMAC4HPO,
        scoring_function,
        tunable_hyperparameters,
        iterations,
        acquisition_function=acquisition.LCB
    )


def smac_smac4hpo_pi(scoring_function, tunable_hyperparameters, iterations):
    """Use SMAC4HPO with Probability of Improvement acquisition function."""
    return _smac_tuning_function(
        SMAC4HPO,
        scoring_function,
        tunable_hyperparameters,
        iterations,
        acquisition_function=acquisition.PI
    )


def smac_hb4ac(scoring_function, tunable_hyperparameters, iterations):
    """Use HyperBand implementation from SMAC."""
    intensifier_kwargs = {
        'initial_budget': 1,
        'max_budget': iterations
    }

    return _smac_tuning_function(
        HB4AC,
        scoring_function,
        tunable_hyperparameters,
        iterations,
        intensifier_kwargs=intensifier_kwargs
    )



# -*- coding: utf-8 -*-

from btb_benchmark.challenges.challenge import Challenge


class Rosenbrock(Challenge):
    """Rosenbrock Challenge.

    This challenge represents the Rosenbrock function, this is a non-convex function, introduced by
    Howard H. Rosenbrock in 1960, which is used as a performance test problem for optimization
    algorithms.[1] It is also known as Rosenbrock's valley or Rosenbrock's banana function.

    The global minimum is inside a long, narrow, parabolic shaped flat valley. To find the valley
    is trivial. To converge to the global minimum, however, is difficult.

    Reference:
        https://en.wikipedia.org/wiki/Rosenbrock_function

    The function is defined by:
        :math:`f(x, y) = (a - x)^2 + b(y - x ^2)^2`

    It has a global minimum at:
        :math:`(x, y) = (a, a^2)` where `f(x, y) = 0`. Usually these parameters are set such that
        :math:`a = 1` and :math:`b = 100`. Only in the trivial case where :math:`a = 0` is the
        function symmetric and the minimum at the origin.

    Args:
        a (int):
            Number that ``a`` will take in the function. Defaults to 1.
        b (int):
            Number that ``b`` will take in the function. Defaults to 100.
        min_x (int):
            Minimum number that the hyperparameter can propouse for ``x``. Defaults to -50.
        max_x (int):
            Maximum number that the hyperparameter can propouse for ``x``. Defaults to 50.
        min_y (int):
            Minimum number that the hyperparameter can propouse for ``y``. Defaults to -50.
        max_y (int):
            Maximum number that the hyperparameter can propouse for ``y``. Defaults to 50.
    """
    def __init__(self, a=1, b=100, min_x=-50, max_x=50, min_y=-50, max_y=50):
        self.a = a
        self.b = b
        self.min_x = min_x
        self.max_x = max_x
        self.min_y = min_y
        self.max_y = max_y

    def get_tunable_hyperparameters(self):
        return {
            'x': {
                'type': 'int',
                'range': [self.min_x, self.max_x],
                'default': None
            },
            'y': {
                'type': 'int',
                'range': [self.min_y, self.max_y],
                'default': None
            }
        }

    def evaluate(self, x, y):
        return -1 * ((self.a - x)**2 + self.b * (y - x**2)**2)



# -*- coding: utf-8 -*-

"""Package where the NumericalHyperParam class and it's childs are defined."""

import sys

import numpy as np

from baytune.tuning.hyperparams.base import BaseHyperParam


class NumericalHyperParam(BaseHyperParam):
    """NumericalHyperParam class.

    The NumericalHyperParam class defines an abstraction to hyperparameters which ranges are
    defined by a numerical value and can take any number within that range.
    """

    dimensions = 1

    def _within_range(self, values, min=0, max=1):
        if (values < min).any() or (values > max).any():
            raise ValueError(
                "Value not within range [{}, {}]: {}".format(min, max, values)
            )


class FloatHyperParam(NumericalHyperParam):
    """FloatHyperParam class.

    The FloatHyperParam class represents a single hyperparameter within a range of ``float``
    numbers, where ``min`` and ``max`` can take as value any float number within that range,
    having ``min`` to be smaller than ``max``.

    Hyperparameter space:
        :math:`h_1, h_2,... h_n` where :math:`h_i = i * (max - min) + min`

    Search space:
        :math:`s_1, s_2,... s_n` where :math:`s_i = (i - min) / (max - min)`

    Args:
        min (float):
            Float number to represent the minimum value that this hyperparameter can take,
            by default is ``None`` which will take the system's minimum float value possible.

        max (float):
            Float number to represent the maximum value that this hyperparameter can take,
            by default is ``None`` which will take the system's maximum float value possible.

        default (float):
            Float number that represents the default value for the hyperparameter. Defaults to
            ``self.min``

        include_min (bool):
            Either or not to include the minimum value in the search space.

        include_max (bool):
            Either or not to include the maximum value in the search space.
    """

    cardinality = np.inf

    def __init__(
        self, min=None, max=None, default=None, include_min=True, include_max=True
    ):
        self.include_min = include_min
        self.include_max = include_max

        if min is None or min == -np.inf:
            min = sys.float_info.min

        if max is None or max == np.inf:
            max = sys.float_info.max

        if min >= max:
            raise ValueError(
                "The ``min`` value can not be greater or equal to ``max`` value."
            )

        if default is None:
            self.default = float(min)
        else:
            self.default = float(default)

        self.min = float(min)
        self.max = float(max)
        self.range = max - min

    def _inverse_transform(self, values):
        """Invert one or more search space values.

        Converts a ``numpy.ndarray`` with normalized values from the search space :math:`s_1,
        s_2,... s_n` to the hyperparameter space of :math:`h_1, h_2,... h_n`. This is
        being computed by multiplying the hyperparameter's range with the values to be denormalized
        and adding the ``self.min`` value to them.

        Args:
            values (numpy.ndarray):
                2D array with values from the search space.

        Returns:
            numpy.ndarray:
                2D ``numpy.ndarray`` containing values from the original hyperparameter space.

        Example:
            The example below shows simple usage case where a FloatHyperParam is being created
            with a range that goes from ``0.1`` to ``0.9`` and it's ``_inverse_transform`` method
            is being called with a valid ``numpy.ndarray`` that contain values from the normalized
            search space and values from the hyperparameter space are being returned.

            >>> instance = FloatHyperParam(min=0.1, max=0.9)
            >>> instance._inverse_transform(np.array([[0.], [1.]]))
            array([[0.1],
                   [0.9]])
            >>> instance._inverse_transform(np.array([[0.875]]))
            array([[0.8]])
        """
        return values * self.range + self.min

    def _transform(self, values):
        """Transform one or more hyperparameter values.

        Converts a ``numpy.ndarray`` with ``float`` values from the original hyperparameter space
        :math:`h_1, h_2,... h_n` into the normalized search space :math:`s_1, s_2,... s_n`.
        This is being computed by substracting the ``self.min`` value from the values to be
        trasnformed and dividing them by the ``self.range`` of the hyperparameter.

        Args:
            values (numpy.ndarray):
                2D array with values from the hyperparameter space to be converted into the
                search space.

        Returns:
            numpy.ndarray:
                2D ``numpy.ndarray`` of shape `(len(values), 1)` containing the search space
                values.

        Example:
            The example below shows simple usage case where a FloatHyperParam is being created
            with a range that goes from ``0.1`` to ``0.9`` and it's ``_transform`` method
            is being called with a valid ``numpy.ndarray`` that contain values from the
            hyperparameter space and values from the search space are being returned.

            >>> instance = FloatHyperParam(min=0.1, max=0.9)
            >>> instance._transform(np.array([[0.1], [0.9]]))
            array([[0.],
                   [1.]])
            >>> instance._transform(np.array([[0.8]]))
            array([[0.875]])
        """
        return (values - self.min) / self.range

    def sample(self, n_samples):
        """Generate sample values in the hyperparameter search space :math:`{0, 1}`.

        Args:
            n_samples (int):
                Number of values to sample.

        Returns:
            numpy.ndarray:
                2D array with shape of `(n_samples, 1)` with normalized values inside the
                search space :math:`{0, 1}`.

        Example:
            The example below shows simple usage case where a FloatHyperParam is being created
            with a range that goes from ``0.1`` to ``0.9`` and it's ``sample`` method
            is being called with a number of samples to be obtained. A ``numpy.ndarray`` with
            values from the search space is being returned.

            >>> instance = FloatHyperParam(min=0.1, max=0.9)
            >>> instance.sample(2)
            array([[0.52058728],
                   [0.00582452]])
        """
        return np.random.random((n_samples, self.dimensions))

    def __repr__(self):
        args = (self.min, self.max, self.default, self.include_min, self.include_max)
        args = "min={}, max={}, default={}, include_min={}, include_max={}".format(
            *args
        )
        return "FloatHyperParam({})".format(args)


class IntHyperParam(NumericalHyperParam):
    """IntHyperParam class.

    The IntHyperParam class represents a single hyperparameter within an range of ``int``
    numbers, where ``min`` and ``max`` can take as value any ``int`` number that compose this range
    having ``min`` to be smaller than ``max``.

    Hyperparameter space:
        :math:`h_1, h_2,... h_n` where :math:`h_i = min + (i - 1) * step`

    Search space:
        :math:`s_1, s_2,... s_n` where :math:`s_i = \\frac{interval}{2} + (i - 1) * interval`

    Args:
        min (int):
            Integer number to represent the minimum value that this hyperparameter can take,
            by default is ``None`` which will take the system's minimum int value possible.

        max (int):
            Integer number to represent the maximum value that this hyperparameter can take,
            by default is ``None`` which will take the system's maximum int value possible.

        default (int):
            Integer number that represents the default value for the hyperparameter. Defaults to
            ``self.min``.

        step (int):
            Increase amount to take for each sample. Defaults to 1.

        include_min (bool):
            Either or not to include the minimum value in the search space.

        include_max (bool):
            Either or not to include the maximum value in the search space.
    """

    dimensions = 1

    def __init__(
        self,
        min=None,
        max=None,
        default=None,
        include_min=True,
        include_max=True,
        step=1,
    ):
        self.include_min = include_min
        self.include_max = include_max

        if min is None or min == -np.inf:
            min = -(sys.maxsize / 2)

        if max is None or max == np.inf:
            max = sys.maxsize / 2

        if min >= max:
            raise ValueError(
                "The `min` value can not be greater or equal to `max` value."
            )

        if default is None:
            self.default = min
        else:
            self.default = int(default)

        self.min = int(min) if include_min else int(min) + 1
        self.max = int(max) if include_max else int(max) - 1
        self.step = step
        self.cardinality = ((self.max - self.min) // step) + 1

        if (self.max - self.min) % self.step:
            raise ValueError(
                "Invalid step of {} for values inside [{}, {}]".format(
                    step, self.min, self.max
                )
            )

        self.interval = self.step / (self.max - self.min + self.step)

    def _inverse_transform(self, values):
        """Invert one or more search space values.

        Converts a ``numpy.ndarray`` with normalized values from the search space :math:`s_1,
        s_2,... s_n` to the hyperparameter space of :math:`h_1, h_2,... h_n`. This is
        being computed by divinding the hyperparameter's interval with the values to be inverted
        and adding the ``self.min`` value to them and resting the 0.5 that has been added during
        the transformation.

        Args:
            values (numpy.ndarray):
                2D array with values from the search space.

        Returns:
            numpy.ndarray:
                2D ``numpy.ndarray`` containing values from the original hyperparameter space.

        Example:
            The example below shows simple usage case where an IntHyperParam is being created
            with a range that goes from ``1`` to ``4`` and it's ``_inverse_transform`` method
            is being called with a valid ``numpy.ndarray`` that contain values from the
            search space and values from the hyperparameter space are being returned.

            >>> instance = IntHyperParam(min=1, max=4)
            >>> instance._inverse_transform(np.array([[0.125], [0.875]]))
            array([[1],
                   [4]])
            >>> instance._inverse_trasnfrom(np.array([[0.625]]))
            array([[3]])
        """
        unscaled_values = values / self.interval - 0.5 + self.min
        rounded = unscaled_values.round()

        # Restrict to make sure that we stay within the valid range
        restricted = np.minimum(np.maximum(rounded, self.min), self.max)

        return restricted.astype(int)

    def _transform(self, values):
        """Transform one or more hyperparameter values.

        Converts a ``numpy.ndarray`` with ``int`` values from the original hyperparameter space
        :math:`h_1, h_2,... h_n` into the normalized search space :math:`s_1, s_2,... s_n`.
        This is being computed by substracting the ``min`` value that the hyperparameter can take
        from the values to be trasnformed and adding them 0.5, then multiply by the interval.

        Args:
            values (numpy.ndarray):
                2D array with values from the hyperparameter space to be converted into the
                search space.

        Returns:
            numpy.ndarray:
                2D ``numpy.ndarray`` of shape `(len(values), 1)` containing the search space
                values.

        Example:
            The example below shows simple usage case where an IntHyperParam is being created
            with a range that goes from ``1`` to ``4`` and it's ``_transform`` method
            is being called with a valid ``numpy.ndarray`` that contain values from the
            hyperparameter space and values from the search space are being returned.

            >>> instance = IntHyperParam(min=1, max=4)
            >>> instance._transform(np.array([[1], [4]]))
            array([[0.125],
                   [0.875]])
            >>> instance._trasnfrom(np.array([[3]]))
            array([[0.625]])
        """
        return ((values - self.min) / self.step + 0.5) * self.interval

    def sample(self, n_samples):
        """Generate sample values in the hyperparameter search space of [0, 1).

        Args:
            n_samples (int):
                Number of values to sample.

        Returns:
            numpy.ndarray:
                2D array with shape of `(n_samples, 1)` with normalized values inside the
                search space :math:`{0, 1}`.

        Example:
            The example below shows simple usage case where a IntHyperParam is being created
            with a range that goes from ``1`` to ``4`` and it's ``sample`` method
            is being called with a number of samples to be obtained. A ``numpy.ndarray`` with
            values from the search space is being returned.

            >>> instance = IntHyperParam(min=1, max=4)
            >>> instance.sample(2)
            array([[0.625],
                   [0.375]])
        """
        sampled = np.random.random((n_samples, self.dimensions))
        inverted = self._inverse_transform(sampled)

        return self._transform(inverted)

    def __repr__(self):
        args = (
            self.min,
            self.max,
            self.default,
            self.include_min,
            self.include_max,
            self.step,
        )
        args = "min={}, max={}, default={}, include_min={}, include_max={}, step={}".format(
            *args
        )
        return "IntHyperParam({})".format(args)



# -*- coding: utf-8 -*-

from baytune.tuning.acquisition.base import BaseAcquisition


class PredictedScoreAcquisition(BaseAcquisition):
    def _acquire(self, candidates, num_candidates=1):
        candidates = candidates if candidates.ndim == 1 else candidates[:, 0]
        return self._get_max_candidates(candidates, num_candidates)



# -*- coding: utf-8 -*-

"""Package where the BaseTuner class and BaseMetaModelTuner are defined."""

import logging
from abc import abstractmethod

import numpy as np

from baytune.tuning.acquisition.base import BaseAcquisition
from baytune.tuning.metamodels.base import BaseMetaModel

LOGGER = logging.getLogger(__name__)


class StopTuning(Exception):
    pass


class BaseTuner:
    """BaseTuner class.

    BaseTuner class is the abstract representation of a tuner that is not based on a model.

    Attributes:
        tunable (baytune.tuning.tunable.Tunable):
            Instance of a tunable class containing hyperparameters to be tuned.
        trials (numpy.ndarray):
            A ``numpy.ndarray`` with shape ``(n, self.tunable.dimensions)`` where ``n`` is the
            number of trials recorded.
        raw_scores (numpy.ndarray):
            A ``numpy.ndarray`` with shape ``(n, 1)`` where ``n`` is the number of scores recorded.
        scores (numpy.ndarray):
            A ``numpy.ndarray`` with shape ``(n, 1)`` where ``n`` is the number of normalized
            scores recorded.

    Args:
        tunable (baytune.tuning.tunable.Tunable):
            Instance of a tunable class containing hyperparameters to be tuned.
        maximize (bool):
            If ``True`` the scores are interpreted as bigger is better, if ``False`` then smaller
            is better. Defaults to ``True``.
    """

    def __init__(self, tunable, maximize=True):
        self.tunable = tunable
        self.trials = np.empty((0, self.tunable.dimensions), dtype=np.float64)
        self._trials_set = set()
        self.raw_scores = np.empty((0, 1), dtype=np.float64)
        self.scores = np.empty((0, 1), dtype=np.float64)
        self.maximize = maximize
        LOGGER.debug(
            ("Creating %s instance with %s hyperparameters and cardinality %s."),
            self.__class__.__name__,
            len(self.tunable.hyperparams),
            self.tunable.cardinality,
        )

    def _check_proposals(self, num_proposals):
        """Validate ``num_proposals`` with ``self.tunable.cardinality`` and ``self.trials``.

        Raises:
            StopTuning:
                A ``StopTuning`` exception is being produced if the amount of requested proposals
                is bigger than the possible combinations and ``allow_duplicates`` is ``False``.
            StopTuning:
                A ``StopTuning`` exception is being produced if the unique amount of recorded
                trials is the same as the amount of combinations available for ``self.tunable``.
            StopTuning:
                A ``StopTuning`` exception is being produced if the unique amount of recorded
                trials is the same as the amount of combinations available for ``self.tunable``.
        """
        if num_proposals > self.tunable.cardinality:
            raise StopTuning(
                "The number of proposals requested is bigger than the combinations: {} of the"
                "``tunable``. Use ``allow_duplicates=True``, if you would like to generate that"
                "amount of combinations.".format(self.tunable.cardinality)
            )

        num_tried = len(self._trials_set)
        if num_tried == self.tunable.cardinality:
            raise StopTuning(
                "All of the possible combinations where recorded. Use ``allow_duplicates=True``"
                "to keep generating combinations."
            )

        if num_tried + num_proposals > self.tunable.cardinality:
            raise StopTuning(
                "The maximum amount of new proposed combinations will exceed the amount of"
                "possible combinations, either use ``num_proposals={}`` to generate the remaining"
                "combinations or ``allow_duplicates=True`` to keep generating more"
                "combinations.".format(self.tunable.cardinality - num_tried)
            )

    def _sample(self, num_proposals, allow_duplicates):
        """Generate a ``numpy.ndarray`` of valid proposals.

        Generates ``num_proposals`` of valid combinations by generating ``proposals`` until
        ``len(valid_proposals) == num_proposals`` different from the ones that have been recorded.

        Args:
            num_proposals (int):
                Amount of proposals to generate.
            allow_duplicates (bool):
                If it's ``False``, the tuner will propose trials that are not recorded. Otherwise
                will generate trials that may have been already recorded.

        Returns:
            numpy.ndarray:
                A ``numpy.ndarray`` with shape ``(num_proposals, self.tunable.dimensions)``.
        """
        if allow_duplicates:
            return self.tunable.sample(num_proposals)

        else:
            valid_proposals = set()

            while len(valid_proposals) < num_proposals:
                proposals = self.tunable.sample(num_proposals)
                proposals = set(map(tuple, proposals))

                valid_proposals.update(proposals - self._trials_set)

            return np.asarray(list(valid_proposals))[:num_proposals]

    @abstractmethod
    def _propose(self, num_proposals, allow_duplicates):
        """Generate ``num_proposals`` number of candidates.

        Args:
            num_proposals (int):
                Number of candidates to create.
            allow_duplicates (bool):
                If it's ``False``, the tuner will propose trials that are not recorded. Otherwise
                will generate trials that can be repeated.

        Returns:
            numpy.ndarray:
                It returns ``numpy.ndarray`` with shape
                ``(num_proposals, len(self.tunable.hyperparameters)``.
        """
        pass

    def propose(self, n=1, allow_duplicates=False):
        """Propose one or more new hyperparameter configurations.

        Validate that the amount of proposals requested is valid when ``allow_duplicates`` is
        ``False`` and raise an exception in case there is any missmatch between ``n``,
        unique ``self.trials`` and ``self.tunable.cardinality``.
        Call the implemented ``_propose`` method and convert the returned data in to hyperparameter
        space values.

        Args:
            n (int):
                Number of candidates to create. Defaults to 1.
            allow_duplicates (bool):
                If it's False, the tuner will propose trials that are not recorded. Otherwise
                will generate trials that can be repeated. Defaults to ``False``.

        Returns:
            dict or list:
                If ``n`` is 1, a ``dict`` will be returned containing the
                hyperparameter names and values. Otherwise, if ``n`` is bigger than 1,
                a list of such dicts is returned.

        Raises:
            ValueError:
                A ``ValueError`` exception is being produced if the amount of requested proposals
                is bigger than the possible combinations and ``allow_duplicates`` is ``False``.
            ValueError:
                A ``ValueError`` exception is being produced if the unique amount of recorded
                trials is the same as the amount of combinations available for ``self.tunable``.
            ValueError:
                A ``ValueError`` exception is being produced if the unique amount of recorded
                trials is the same as the amount of combinations available for ``self.tunable``.

        Example:
            The example below shows simple usage case where an ``UniformTuner`` is being imported,
            instantiated with a ``tunable`` object and it's method propose is being called
            three times, first with a single proposal, a second with two proposals forcing them to
            be different and once where the values can be repeated.

            >>> from baytune.tuning.tunable import Tunable
            >>> from baytune.tuning.hyperparams import BooleanHyperParam
            >>> from baytune.tuning.hyperparams import CategoricalHyperParam
            >>> from baytune.tuning.tuners import UniformTuner
            >>> bhp = BooleanHyperParam()
            >>> chp = CategoricalHyperParam(['cat', 'dog'])
            >>> tunable = Tunable({'bhp': bhp, 'chp': chp})
            >>> tuner = UniformTuner(tunable)
            >>> tuner.propose(1)
            {'bhp': True, 'chp': 'dog'}
            >>> tuner.propose(2)
            [{'bhp': True, 'chp': 'cat'}, {'bhp': True, 'chp': 'dog'}]
            >>> tuner.propose(2, allow_duplicates=True)
            [{'bhp': False, 'chp': 'dog'}, {'bhp': False, 'chp': 'dog'}]
        """

        if not allow_duplicates:
            self._check_proposals(n)

        proposed = self._propose(n, allow_duplicates)

        hyperparameters = self.tunable.inverse_transform(proposed)
        hyperparameters = hyperparameters.to_dict(orient="records")

        if n == 1:
            hyperparameters = hyperparameters[0]

        return hyperparameters

    def record(self, trials, scores):
        """Record one or more ``trials`` with the associated ``scores``.

        ``Trials`` are recorded with their associated ``scores``. The amount of trials
        must be equal to the amount of scores recived and vice versa.

        Args:
            trials (pandas.DataFrame, pandas.Series, dict, list(dict), 2D array-like):
                Values of shape ``(n, len(self.tunable.hyperparameters))`` or dict with keys that
                are ``self.tunable.names``.

            scores (single value or array-like):
                A single value or array-like of values representing the score achieved with the
                trials.

        Raises:
            ValueError:
                A ``ValueError`` exception is being produced if ``len(trials)`` is not equal to
                ``len(scores)``.

        Example:
            The example below shows simple usage case where an ``UniformTuner`` is being imported,
            instantiated with a ``tunable`` object and it's method record is being called two times
            with valid trials and scores.

            >>> from baytune.tuning.tunable import Tunable
            >>> from baytune.tuning.hyperparams import BooleanHyperParam
            >>> from baytune.tuning.hyperparams import CategoricalHyperParam
            >>> from baytune.tuning.tuners import UniformTuner
            >>> bhp = BooleanHyperParam()
            >>> chp = CategoricalHyperParam(['cat', 'dog'])
            >>> tunable = Tunable({'bhp': bhp, 'chp': chp})
            >>> tuner = UniformTuner(tunable)
            >>> tuner.record({'bhp': True, 'chp': 'cat'}, 0.8)
            >>> trials = [{'bhp': False, 'chp': 'cat'}, {'bhp': True, 'chp': 'dog'}]
            >>> scores = [0.8, 0.1]
            >>> tuner.record(trials, scores)
        """

        trials = self.tunable.transform(trials)
        scores = scores if isinstance(scores, (list, np.ndarray)) else [scores]

        if len(trials) != len(scores):
            raise ValueError(
                "The amount of trials must be equal to the amount of scores."
            )

        self.trials = np.append(self.trials, trials, axis=0)
        self._trials_set.update(map(tuple, trials))
        self.raw_scores = np.append(self.raw_scores, scores)
        self.scores = self.raw_scores if self.maximize else -self.raw_scores

    def __str__(self):
        return (
            "{}\n" "  hyperparameters: {}\n" "  dimensions: {}\n" "  cardinality: {}"
        ).format(
            self.__class__.__name__,
            len(self.tunable.hyperparams),
            self.tunable.dimensions,
            self.tunable.cardinality,
        )


class BaseMetaModelTuner(BaseTuner, BaseMetaModel, BaseAcquisition):
    """BaseMetaModelTuner class.

    BaseMetaModelTuner class is the abstract representation of a tuner that is based
    on a model and an ``Acquisition``. This model will try to `predict` the
    score that will be obtained with the proposed parameters by being trained
    over the ``self.trials`` and ``self.raw_scores`` recorded by the user.

    Attributes:
        tunable (baytune.tuning.tunable.Tunable):
            Instance of a tunable class containing hyperparameters to be tuned.
        trials (numpy.ndarray):
            A ``numpy.ndarray`` with shape ``(n, self.tunable.dimensions)`` where ``n`` is the
            number of trials recorded.
        scores (numpy.ndarray):
            A ``numpy.ndarray`` with shape ``(n, 1)`` where ``n`` is the number of scores recorded.

    Args:
        tunable (baytune.tuning.tunable.Tunable):
            Instance of a tunable class containing hyperparameters to be tuned.
        num_candidates (int):
            Number of samples to generate and select the best of it for each proposal. Defaults to
            1000.
        maximize (bool):
            If ``True`` the model will understand that the score bigger is better, if ``False``
            the smaller is better. Defaults to ``True``.
        min_trials (int):
            Number of recorded ``trials`` needed to perform a fitting over the model.
            Defaults to 5.
    """

    _metamodel_kwargs = None
    _acquisition_kwargs = None

    def __init__(self, tunable, maximize=True, num_candidates=1000, min_trials=5):
        self.num_candidates = num_candidates
        self.min_trials = min_trials
        super().__init__(tunable, maximize)
        self.__init_metamodel__(**(self._metamodel_kwargs or dict()))
        self.__init_acquisition__(**(self._acquisition_kwargs or dict()))

    def _propose(self, num_proposals, allow_duplicates):
        if self.min_trials > len(self._trials_set):
            LOGGER.debug(
                "Not enough samples recorded to generate predictions, "
                "generating random proposal."
            )
            return self._sample(num_proposals, allow_duplicates)

        num_samples = num_proposals * self.num_candidates
        if not allow_duplicates:
            remaining = self.tunable.cardinality - len(self._trials_set)
            num_samples = min(remaining, num_samples)

        proposals = self._sample(num_samples, allow_duplicates)

        predicted = self._predict(proposals)
        index = self._acquire(predicted, num_proposals)

        return proposals[index]

    def record(self, trials, scores):
        """Record one or more ``trials`` with the associated ``scores`` and re-fit the model.

        ``Trials`` are recorded with the associated ``scores`` to them. The amount of trials
        must be equal to the amount of scores recived and vice versa. Once recorded, the ``model``
        is being fitted with ``self.trials`` and ``self.raw_scores`` that contain any previous
        records and the ones that where just recorded.

        Args:
            trials (pandas.DataFrame, pandas.Series, dict, list(dict), 2D array-like):
                Values of shape ``(n, len(self.tunable.hyperparameters))`` or dict with keys that
                are ``self.tunable.names``.

            scores (single value or array-like):
                A single value or array-like of values representing the score achieved with the
                trials.

        Raises:
            ValueError:
                A ``ValueError`` exception is being produced if ``len(trials)`` is not equal to
                ``len(scores)``.

        Example:
            The example below shows simple usage case where an ``UniformTuner`` is being imported,
            instantiated with a ``tunable`` object and it's method record is being called two times
            with valid trials and scores.

            >>> from baytune.tuning.tunable import Tunable
            >>> from baytune.tuning.hyperparams import BooleanHyperParam
            >>> from baytune.tuning.hyperparams import CategoricalHyperParam
            >>> from baytune.tuning.tuners import UniformTuner
            >>> bhp = BooleanHyperParam()
            >>> chp = CategoricalHyperParam(['cat', 'dog'])
            >>> tunable = Tunable({'bhp': bhp, 'chp': chp})
            >>> tuner = UniformTuner(tunable)
            >>> tuner.record({'bhp': True, 'chp': 'cat'}, 0.8)
            >>> trials = [{'bhp': False, 'chp': 'cat'}, {'bhp': True, 'chp': 'dog'}]
            >>> scores = [0.8, 0.1]
            >>> tuner.record(trials, scores)
        """
        super().record(trials, scores)
        if len(self.trials) >= self.min_trials:
            LOGGER.debug("Fitting the model with %s samples." % len(self.trials))
            self._fit(self.trials, self.scores)
